/*
 *  linux/fs/proc/proc_ub.c
 *
 *  Copyright (C)  1998-2000  Andrey V. Savochkin <saw@saw.sw.com.sg>
 *
 * TODO:
 *
 * Changes:
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>

#include <linux/beancounter.h>
#include <linux/ubhash.h>

#include <asm/page.h>
#include <asm/uaccess.h>

/*
 *	Debugging option bits
 *  1	single file read debug
 */
#define UB_DEBUG 0


#if BITS_PER_LONG == 32
#define UB_PROC_LINE_SHIFT	6
#define UB_PROC_LINE_TEXT	(5+2+11+1+10+1+10+1+10+1+10)
#else
#define UB_PROC_LINE_SHIFT	7
#define UB_PROC_LINE_TEXT	(10+2+11+1+20+1+20+1+20+1+20)
#endif
#define UB_PROC_LINE_LEN	(1 << UB_PROC_LINE_SHIFT)
#define UB_PROC_LINE_SPACES	(UB_PROC_LINE_LEN - UB_PROC_LINE_TEXT - 1)

static const char *ub_rnames[UB_RESOURCES] = {
	"kmemsize",
	"lockedpages",
	"totvmpages",
	"ipcshmpages",
	"anonshpages",
	"numproc",
	"rsspages",
	"vmspaceguar",
	"oomguar",
	"numsock",
	"numflock",
	"numpty",
	"numsiginfo",
};

static void out_proc_head(char *buf)
{
#if BITS_PER_LONG == 32
	sprintf(buf, "  uid  resource          held        max    barrier"
		     "      limit");
#else
	sprintf(buf, "       uid  resource                    held"
		     "                  max              barrier"
		     "                limit");
#endif
	memset(buf + UB_PROC_LINE_TEXT,	' ', UB_PROC_LINE_SPACES);
	buf[UB_PROC_LINE_LEN - 1] = '\n';
}

static void out_proc_beancounter(char *buf, struct user_beancounter *ub, int r)
{
	if (!r)
#if BITS_PER_LONG == 32
		sprintf(buf, "%5u: %-11s %10lu %10lu %10lu %10lu",
			(unsigned)ub->ub_uid, ub_rnames[r],
			ub->ub_held[r], ub->ub_maxheld[r], 
			ub->ub_barrier[r], ub->ub_limit[r]);
#else
		sprintf(buf, "%10u: %-11s %20lu %20lu %20lu %20lu",
			(unsigned)ub->ub_uid, ub_rnames[r],
			ub->ub_held[r], ub->ub_maxheld[r], 
			ub->ub_barrier[r], ub->ub_limit[r]);
#endif
	else
#if BITS_PER_LONG == 32
		sprintf(buf, "       %-11s %10lu %10lu %10lu %10lu",
			ub_rnames[r],
			ub->ub_held[r], ub->ub_maxheld[r], 
			ub->ub_barrier[r], ub->ub_limit[r]);
#else
		sprintf(buf, "            %-11s %20lu %20lu %20lu %20lu",
			ub_rnames[r],
			ub->ub_held[r], ub->ub_maxheld[r], 
			ub->ub_barrier[r], ub->ub_limit[r]);
#endif
	memset(buf + UB_PROC_LINE_TEXT,	' ', UB_PROC_LINE_SPACES);
	buf[UB_PROC_LINE_LEN - 1] = '\n';
}

static ssize_t ub_proc_read(struct file *file, char *usrbuf, size_t len,
		loff_t *poff)
{
	ssize_t retval;
	char *buf;
	unsigned long flags;
	int i, resource;
	struct ub_hash_slot *slot;
	struct user_beancounter *ub;
	size_t n;
	int rem, produced, job, tocopy;
	const int is_capable = is_super_ve(current) && 
		(capable(CAP_DAC_OVERRIDE) || capable(CAP_DAC_READ_SEARCH));
	
	retval = -ENOBUFS;
	buf = (char *)__get_free_page(GFP_KERNEL);
	if (buf == NULL)
		goto out;

	retval = 0;
	if (!is_capable && current->luid == (uid_t)-1)
		goto out_free;
	if (*poff < 0)
		goto out_free;
again:
	i = 0;
	slot = ub_hash;
	n = *poff >> UB_PROC_LINE_SHIFT;	/* in lines */
	rem = *poff & (UB_PROC_LINE_LEN - 1);	/* in bytes */
	produced = 0;
	if (!n) {
		out_proc_head(buf);
		produced += UB_PROC_LINE_LEN;
		n++;
	}
	n--;
	while (1) {
		lock_beancounters(slot, flags);
		for (ub = slot->ubh_beans; ub != NULL && n >= UB_RESOURCES;
				ub = ub->ub_next)
			if (is_capable || current->luid == ub->ub_uid)
				n -= UB_RESOURCES;
		if (ub != NULL)
			break;
		unlock_beancounters(slot, flags);
		if (++i >= UB_HASH_SIZE)
			goto out_free;
		++slot;
	}
	rem += n << UB_PROC_LINE_SHIFT;
	job = PAGE_SIZE;
	if (len < PAGE_SIZE - rem)
		job = rem + len;
	while (produced < job) {
		if (is_capable || (current->luid == ub->ub_uid))
			for (resource = 0; produced < job && resource < UB_RESOURCES;
					resource++, produced += UB_PROC_LINE_LEN)
				out_proc_beancounter(buf + produced, ub, resource);
		if (produced >= job)
			break;
		ub = ub->ub_next;
checkub:
		if (ub != NULL)
			continue;
		if (++i >= UB_HASH_SIZE)
			break;
		unlock_beancounters(slot, flags);
		++slot;
		lock_beancounters(slot, flags);
		ub = slot->ubh_beans;
		goto checkub;
	}
	unlock_beancounters(slot, flags);
#if UB_DEBUG & 1
	printk(KERN_DEBUG "UB_PROC: produced %d, job %d, rem %d\n",
			produced, job, rem);
#endif
	if (produced <= rem)
		goto out_free;
	/* Temporay buffer `buf' contains `produced' bytes.
	 * Extract no more than `len' bytes at offset `rem'.
	 */
	tocopy = produced - rem;
	if (len < tocopy)
		tocopy = len;
	if (!tocopy)
		goto out_free;
	if (copy_to_user(usrbuf, buf + rem, tocopy))
		goto fail;
	*poff += tocopy;
	len -= tocopy;
	retval += tocopy;
	if (!len)
		goto out_free;
	usrbuf += tocopy;
	goto again;
fail:
	retval = -EFAULT;
out_free:
	free_page((unsigned long)buf);
out:
	return retval;
}

struct file_operations ub_file_operations = {
	read: ub_proc_read,
};
