/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#ifdef __KERNEL__

#include <linux/sched.h>
#include <linux/pagemap.h>
#include <linux/reiserfs_fs.h>
#include <linux/locks.h>

#else

#include "nokernel.h"

#endif


/* access to tail : when one is going to read tail it must make sure, that is not running.
 direct2indirect and indirect2direct can not run concurrently */


/* Converts direct items to an unformatted node. Panics if file has no
   tail. -ENOSPC if no disk space for conversion */
/* path points to first direct item of the file regarless of how many of
   them are there */
int direct2indirect (struct reiserfs_transaction_handle *th, struct inode * inode, 
		     struct path * path, struct buffer_head * unbh,
		     loff_t tail_offset)
{
    struct super_block * sb = inode->i_sb;
    struct buffer_head *up_to_date_bh ;
    struct item_head * p_le_ih = PATH_PITEM_HEAD (path);
    struct cpu_key end_key;  /* Key to search for the last byte of the
				converted item. */
    struct item_head ind_ih; /* new indirect item to be inserted or
                                key of unfm pointer to be pasted */
    int	n_blk_size,
      n_retval;	  /* returned value for reiserfs_insert_item and clones */
    struct unfm_nodeinfo unfm_ptr;  /* Handle on an unformatted node
				       that will be inserted in the
				       tree. */


    sb->u.reiserfs_sb.s_direct2indirect ++;

    n_blk_size = sb->s_blocksize;

    /* and key to search for append or insert pointer to the new
       unformatted node. */
    copy_item_head (&ind_ih, p_le_ih);
    set_le_ih_k_offset (&ind_ih, tail_offset);
    set_le_ih_k_type (&ind_ih, TYPE_INDIRECT);

    /* Set the key to search for the place for new unfm pointer */
    make_cpu_key (&end_key, inode, tail_offset, TYPE_INDIRECT, 4);

    // FIXME: we could avoid this 
    if ( search_for_position_by_key (sb, &end_key, path) == POSITION_FOUND )
	reiserfs_panic (sb, "PAP-14030: direct2indirect: "
			"pasted or inserted byte exists in the tree");
    
    p_le_ih = PATH_PITEM_HEAD (path);

    unfm_ptr.unfm_nodenum = cpu_to_le32 (unbh->b_blocknr);
    unfm_ptr.unfm_freespace = 0; // ???
    
    if ( is_statdata_le_ih (p_le_ih) )  {
	/* Insert new indirect item. */
	set_ih_free_space (&ind_ih, 0); /* delete at nearest future */
	ind_ih.ih_item_len = cpu_to_le16 (UNFM_P_SIZE);
	PATH_LAST_POSITION (path)++;
	n_retval = reiserfs_insert_item (th, path, &end_key, &ind_ih, 
					 (char *)&unfm_ptr);
    } else {
	/* Paste into last indirect item of an object. */
	n_retval = reiserfs_paste_into_item(th, path, &end_key,
					    (char *)&unfm_ptr, UNFM_P_SIZE);
    }
    if ( n_retval ) {
	return n_retval;
    }

    // note: from here there are two keys which have matching first
    // three key components. They only differ by the fourth one.


    /* Set the key to search for the direct items of the file */
    make_cpu_key (&end_key, inode, max_reiserfs_offset (inode), TYPE_DIRECT, 4);

    /* Move bytes from the direct items to the new unformatted node
       and delete them. */
    while (1)  {
	int item_len, first_direct;

	/* end_key.k_offset is set so, that we will always have found
           last item of the file */
	if ( search_for_position_by_key (sb, &end_key, path) == POSITION_FOUND )
	    reiserfs_panic (sb, "PAP-14050: direct2indirect: "
			    "direct item (%k) not found", &end_key);
	p_le_ih = PATH_PITEM_HEAD (path);
#ifdef CONFIG_REISERFS_CHECK
	if (!is_direct_le_ih (p_le_ih))
	    reiserfs_panic (sb, "vs-14055: direct2indirect: "
			    "direct item expected, found %h", p_le_ih);
#endif
	if ((le_ih_k_offset (p_le_ih) & (n_blk_size - 1)) == 1)
	    first_direct = 1;
	else
	    first_direct = 0;
	item_len = le16_to_cpu (p_le_ih->ih_item_len);

	/* we only send the unbh pointer if the buffer is not up to date.
	** this avoids overwriting good data from writepage() with old data
	** from the disk or buffer cache
	*/
	if (buffer_uptodate(unbh) || Page_Uptodate(unbh->b_page)) {
	    up_to_date_bh = NULL ;
	} else {
	    up_to_date_bh = unbh ;
	}
	n_retval = reiserfs_delete_item (th, path, &end_key, inode, 
	                                 up_to_date_bh) ;

	if (first_direct && item_len == n_retval)
	    // done: file does not have direct items anymore
	    break;

    }

				/* hardcoding is bad, especially
without commenting.  Isn't this defined using some nice mnemonic
somewhere, like LAST_KEY or some such? */
    inode->u.reiserfs_i.i_first_direct_byte = U32_MAX;

    return 0;
}


/* stolen from fs/buffer.c */
void reiserfs_unmap_buffer(struct buffer_head *bh) {
  if (buffer_mapped(bh)) {
    if (buffer_journaled(bh) || buffer_journal_dirty(bh)) {
      BUG() ;
    }
    mark_buffer_clean(bh) ;
    wait_on_buffer(bh) ;
    // clear_bit(BH_Uptodate, &bh->b_state) ;
    clear_bit(BH_Mapped, &bh->b_state) ;
    clear_bit(BH_Req, &bh->b_state) ;
    clear_bit(BH_New, &bh->b_state) ;
  }
}

/* unmap all the buffers on this page.  This makes sure reiserfs_get_block
** will be called by the address space operations.  It should be used
** after packing the tail to invalidate the page cache version of the
** tail as an unformatted node.
*/ 
static void unmap_buffers(struct page *page) {
  struct buffer_head *bh ;
  struct buffer_head *head ;
  struct buffer_head *next ;

  if (page) {
    if (page->buffers) {
      head = page->buffers ;
      bh = head ;
      do {
	next = bh->b_this_page ;
	reiserfs_unmap_buffer(bh) ;
	bh = next ;
      } while (bh != head) ;
    }
  } 
}

/*
** This must be called with i_sem held, and the page locked.  Holding
** i_sem prevents anyone from changing the file size while we are converting
** the tail, and having the page locked prevents readpage/writepage from
** trying to use the tail while we are converting it.
*/
int indirect2direct (struct reiserfs_transaction_handle *th, 
		     struct inode * p_s_inode,
		     struct page *page, 
		     struct path * p_s_path, /* path to the indirect item. */
		     struct cpu_key * p_s_item_key, /* Key to look for unformatted node pointer to be cut. */
		     loff_t n_new_file_size, /* New file size. */
		     char * p_c_mode)
{
    struct super_block * p_s_sb = p_s_inode->i_sb;
    struct item_head      s_ih;
    unsigned long n_block_size = p_s_sb->s_blocksize;
    char * tail;
    int tail_len, round_tail_len;
    loff_t pos, pos1; 
    struct cpu_key key;

				/* What does this do? -Hans */
			  /* it increments s_indirect2direct by one --Chris */
    p_s_sb->u.reiserfs_sb.s_indirect2direct ++;

    /* default mode is to skip the balance.  If the conversion works, we
    ** set the mode to M_CUT later on
    */
    *p_c_mode = M_SKIP_BALANCING;

    /* store item head path points to. */
    copy_item_head (&s_ih, PATH_PITEM_HEAD(p_s_path));

				/* maintenance note: like the quota
                                   code will also, this will break if
                                   anyone ever implements tails for
                                   holes */
    tail_len = (n_new_file_size & (n_block_size - 1));
				/* comment anyone? I suspect this is
                                   some sort of badly done alpha port
                                   code, am I right? */
    if (!old_format_only (p_s_sb))
	round_tail_len = ROUND_UP (tail_len);
    else
	round_tail_len = tail_len;

    /* calculate the byte offset in the file for the tail */
    pos = le_ih_k_offset (&s_ih) - 1 + (le16_to_cpu (s_ih.ih_item_len) / UNFM_P_SIZE - 1) * p_s_sb->s_blocksize;
    pos1 = pos;

    tail = (char *)kmap(page) ; /* this can schedule */

				/* while it cannot disappear, it can
                                   be rebalanced. So here we check for
                                   that.  -Hans */
    if (path_changed (&s_ih, p_s_path)) {
	/* re-search indirect item */
	if ( search_for_position_by_key (p_s_sb, p_s_item_key, p_s_path) == POSITION_NOT_FOUND )
	    reiserfs_panic(p_s_sb, "PAP-5520: indirect2direct: "
			   "item to be converted %k does not exist", p_s_item_key);
	copy_item_head(&s_ih, PATH_PITEM_HEAD(p_s_path));
#ifdef CONFIG_REISERFS_CHECK
	pos = le_ih_k_offset (&s_ih) - 1 + 
	    (le16_to_cpu (s_ih.ih_item_len) / UNFM_P_SIZE - 1) * p_s_sb->s_blocksize;
	if (pos != pos1)
	    reiserfs_panic (p_s_sb, "vs-5530: indirect2direct: "
			    "tail position changed while we were reading it");
#endif
    }


    /* Set direct item header to insert. */
    make_le_item_head (&s_ih, 0, inode_items_version (p_s_inode), pos1 + 1,
		       TYPE_DIRECT, round_tail_len, 0xffff/*ih_free_space*/);
				/* how does this work when buffers
                                   are not equal to pages in size? */
    tail = tail + (pos & (PAGE_CACHE_SIZE - 1)) ;

				/* this does what, tell the path to point to the position we are to insert? */
    PATH_LAST_POSITION(p_s_path)++;

    key = *p_s_item_key;
    set_cpu_key_k_type (&key, TYPE_DIRECT);
    key.key_length = 4;
    /* Insert tail as new direct item in the tree */
    if ( reiserfs_insert_item(th, p_s_path, &key, &s_ih,
			      tail ? tail : NULL) < 0 ) {
	/* No disk memory. So we can not convert last unformatted node
	   to the direct item.  In this case we used to adjust
	   indirect items's ih_free_space. Now ih_free_space is not
	   used, it would be ideal to write zeros to corresponding
	   unformatted node. For now i_size is considered as guard for
	   going out of file size */
	kunmap(tail) ;
	return n_block_size - round_tail_len;
    }
    kunmap(tail) ;

    /* now that we have converted the tail, the last buffer in this
    ** file is not valid for writing (the data has moved elsewhere).
    ** So, we unmap it to force the address space operations to call
    ** get_block instead of using the cached data.
    **
    ** If buffer size < page size, this might unmap the buffer before
    ** the end of the file.  This is a bug since it cleans the buffer
    ** and might lose dirty data.  FIXME
    */
    unmap_buffers(page) ;

    // note: we have now the same as in above direct2indirect
    // conversion: there are two keys which have matching first three
    // key components. They only differ by the fouhth one.

    /* We have inserted new direct item and must remove last
       unformatted node. */
    p_s_inode->i_blocks += (p_s_sb->s_blocksize / 512);
    *p_c_mode = M_CUT;

    /* we store position of first direct item in the in-core inode */
    //mark_file_with_tail (p_s_inode, pos1 + 1);
    p_s_inode->u.reiserfs_i.i_first_direct_byte = pos1 + 1;

    return n_block_size - round_tail_len;
}



