#ifndef _LINUX_BEANCOUNTER_H
#define _LINUX_BEANCOUNTER_H

/*
 *	Resource list.
 */

#define UB_KMEMSIZE	0	/* Unswappable kernel memory size including
					   struct task, page directories, etc.
				   Still unimplemeted are:
					   socket buffers, files, temporary
					   space for poll, siginfo
				 */
#define UB_LOCKEDPAGES	1	/* Mlock()ed pages. */
#define UB_TOTVMPAGES	2	/* Address space size in pages. */
#define UB_SHMPAGES	3	/* IPC SHM segment size. */
#define UB_ZSHMPAGES	4	/* Anonymous shared memory. */
#define UB_NUMPROC	5	/* Number of processes. */
#define UB_RESPAGES	6	/* All resident pages, for swapout guarantee. */
#define UB_SPCGUARPAGES	7	/* Guarantees for address space allocations.
				   Only barier is used, no accounting.
				 */
#define UB_OOMGUARPAGES	8	/* Guarantees against OOM kill.
				   Only limit is used, no accounting.
				 */
#define UB_NUMSOCK	9	/* Number of sockets. */
#define UB_NUMFLOCK	10	/* Number of file locks. */
#define UB_NUMPTY	11	/* Number of PTYs. */
#define UB_NUMSIGINFO	12	/* Number of siginfos. */
#define UB_RESOURCES	13


#ifdef __KERNEL__

#include <asm/atomic.h>
#include <linux/spinlock.h>
 
/*
 * UB_MAXVALUE is essentially LONG_MAX declared in a cross-compiling safe form.
 */
#define UB_MAXVALUE	( (1UL << (sizeof(unsigned long)*8-1)) - 1)


/*
 *	Resource management structures
 * Serialization issues:
 *   beancounter list management is protected via per hash entry lock
 *   task pointers and luid are set only for current task and only once
 *   refcount is managed atomically
 *   value and limit comparison and change are protected by per-node spinlock
 */
 
struct user_beancounter
{
	atomic_t		ub_refcount;
	struct			user_beancounter *ub_next;
	spinlock_t		ub_lock;
	uid_t			ub_uid;
	/* dynamic swap-out priority */
	signed long		ub_swp_pri;
	/* total weight of resident pages, see mm/kubd.c */
	unsigned long long	ub_held_pages;
	/* consumed resources */
	unsigned long		ub_held[UB_RESOURCES];
	/* maximum amount of consumed resources through the whole lifetime */
	unsigned long	ub_maxheld[UB_RESOURCES]; /* Max held resources. */
	/* A barrier over which resource allocations are failed gracefully.
	 * If the amount of consumed memory is over the barrier further sbrk()
	 * or mmap() calls fail, the existing processes are not killed. */
	unsigned long		ub_barrier[UB_RESOURCES];
	/* hard resource limit */
	unsigned long		ub_limit[UB_RESOURCES];
};


#ifndef CONFIG_USER_RESOURCE

extern inline void get_beancounter(struct user_beancounter *ub) {;}
extern inline void put_beancounter(struct user_beancounter *ub) {;}

#define UB_DECLARE_CHARGE(name, dclargs, realargs)	\
extern inline int charge_##name dclargs	{ return 0; }
#define UB_DECLARE_UNCHARGE(name, dclargs, realargs)	\
extern inline void uncharge_##name dclargs { ; }

#else

/*
 * NULL beancounter is handled separately for better performance of tasks
 * without a beancounter.  But I'm not sure how much it gains in real life.
 * --SAW
 */

extern void __put_beancounter(struct user_beancounter *ub);
 
extern struct task_struct *select_worst_task(void);

extern inline void put_beancounter(struct user_beancounter *ub)
{
	if (ub == NULL)
		return;
	__put_beancounter(ub);
}

/*
 *	Create a new beancounter reference
 */
 
extern inline void get_beancounter(struct user_beancounter *ub)
{
	if (ub == NULL)
		return;
#if 0
	printk("get beancounter %p for %.20s pid %d\n", ub, current->comm, current->pid);
#endif
	atomic_inc(&ub->ub_refcount);
}

#define UB_DECLARE_CHARGE(name, dclargs, realargs) \
extern int __charge_##name dclargs;		\
						\
extern inline int charge_##name dclargs		\
{						\
	if (ub == NULL)				\
		return 0;			\
	return __charge_##name realargs;	\
}

#define	UB_DECLARE_UNCHARGE(name, dclargs, realargs) \
extern void __uncharge_##name dclargs;		\
						\
extern inline void uncharge_##name dclargs	\
{						\
	if (ub == NULL)				\
		return;				\
	__uncharge_##name realargs;		\
}

#endif /* CONFIG_USER_RESOURCE */


/*
 *	Resource charging
 * Change user's account and compare against limits
 */

UB_DECLARE_CHARGE(task, (struct user_beancounter *ub), (ub))
UB_DECLARE_UNCHARGE(task, (struct user_beancounter *ub), (ub))
UB_DECLARE_CHARGE(memory,
	(struct user_beancounter *ub, unsigned long size, unsigned vm_flags,
	 int strict),
	(ub, size, vm_flags, strict))
UB_DECLARE_UNCHARGE(memory,
	(struct user_beancounter *ub, unsigned long size, unsigned vm_flags),
	(ub, size, vm_flags))
UB_DECLARE_CHARGE(locked_mem,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))
UB_DECLARE_UNCHARGE(locked_mem,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))
UB_DECLARE_CHARGE(kmem,
	(struct user_beancounter *ub, unsigned long size, int strict),
	(ub, size, strict))
UB_DECLARE_UNCHARGE(kmem,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))
UB_DECLARE_CHARGE(shmpages,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))
UB_DECLARE_UNCHARGE(shmpages,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))
UB_DECLARE_CHARGE(sock,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))
UB_DECLARE_UNCHARGE(sock,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))
UB_DECLARE_CHARGE(flock,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))	
UB_DECLARE_UNCHARGE(flock,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))	
UB_DECLARE_CHARGE(pty,
	(struct user_beancounter *ub),
	(ub))
UB_DECLARE_UNCHARGE(pty,
	(struct user_beancounter *ub),
	(ub))	
UB_DECLARE_CHARGE(siginfo,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))	
UB_DECLARE_UNCHARGE(siginfo,
	(struct user_beancounter *ub, unsigned long size),
	(ub, size))		

#undef UB_DECLARE

#endif /* __KERNEL__ */
#endif /* _LINUX_BEANCOUNTER_H */
