/*
 *  linux/kernel/beancounter.c
 *
 *  Copyright (C) 1998  Alan Cox
 *                1998-2000  Andrey V. Savochkin <saw@saw.sw.com.sg>
 *
 * TODO:
 *   - more intelligent limit check in mremap(): currently the new size is
 *     charged and _then_ old size is uncharged
 *   - limits on number of file descriptors
 *   - problem: bad pmd page handling
 *   - think about unserialized accesses to guarantee fields
 *   - sizeof(struct inode) and omem for sockets
 *   - think about sizeof(struct inode) charge in general and limits for number
 *     of files
 *   - add missing get/put_beancounter for IPC
 *   - charge for UNIX domain sockets
 *
 * Changes:
 *   1999/08/17  Marcelo Tosatti <marcelo@conectiva.com.br>
 *	- Set "barrier" and "limit" parts of limits atomically.
 *   1999/10/06  Marcelo Tosatti <marcelo@conectiva.com.br>
 *	- setublimit system call.
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>

#include <linux/beancounter.h>
#include <linux/ubhash.h>

#include <linux/spinlock.h>
#include <asm/uaccess.h>
#include <asm/processor.h>

#ifdef CONFIG_USER_RESOURCE_PROC
#include <linux/proc_fs.h>
#endif

/*
 *	Various debugging stuff
 *  1	creation/destruction
 *  2	charge operations
 *  4	resource limit change
 */
#define UB_DEBUG 0



#ifdef CONFIG_USER_RESOURCE



extern int max_threads;

static kmem_cache_t *ub_cachep;

static struct user_beancounter default_beancounter;

/*
 *	Per user resource beancounting. Resources are tied to their 
 *	luid. The resource structure itself is tagged both to the process
 *	and the charging resources (a socket doesnt want to have to search
 *	for things at irq time for example). Reference counters keep things
 *	in hand.
 *
 *	The case where a user creates resource, kills all his processes and
 *	then starts new ones is correctly handled this way. The refcounters
 *	will mean the old entry is still around with resource tied to it.
 */
static struct user_beancounter *get_beancounter_byuid(uid_t uid, int create)
{
	struct user_beancounter *ub, *walkp;
	unsigned long flags;
	struct ub_hash_slot *slot = &ub_hash[ub_hash_fun(uid)];
	
	ub = (struct user_beancounter *)kmem_cache_alloc(ub_cachep, GFP_KERNEL);
	if (ub == NULL)
		return NULL;

	lock_beancounters(slot, flags);

	walkp = slot->ubh_beans;
	while (walkp != NULL && walkp->ub_uid != uid)
		walkp = walkp->ub_next;
	
	if (walkp == NULL) {
		if (create) {
#if UB_DEBUG & 1
			printk(KERN_DEBUG "Creating ub %p in slot %p\n", ub, slot);
#endif
			memcpy(ub, &default_beancounter, sizeof(*ub));
			ub->ub_next = slot->ubh_beans;
			slot->ubh_beans = ub;
			ub->ub_uid = uid;
			walkp = ub;
			ub = NULL;
		}
		/* if create isn't set, just return NULL */
	} else {
		/* beancounter already exists */
		atomic_inc(&walkp->ub_refcount);
	}

	if (ub != NULL)
		kmem_cache_free(ub_cachep, ub);
	unlock_beancounters(slot, flags);
	return walkp;
}


void __put_beancounter(struct user_beancounter *ub)
{
	struct user_beancounter **ubptr;
	unsigned long flags;
	struct ub_hash_slot *slot = &ub_hash[ub_hash_fun(ub->ub_uid)];
	
#if UB_DEBUG & 1
	printk(KERN_DEBUG "__put bc %p (cnt %d) for %.20s pid %d cur %08lx cpu %d.\n",
			ub, atomic_read(&ub->ub_refcount),
			current->comm, current->pid,
			(unsigned long)current, smp_processor_id());
#endif
	lock_beancounters(slot, flags);

	if (!atomic_dec_and_test(&ub->ub_refcount)) {
		unlock_beancounters(slot, flags);
		return;
	}

	/*
	 *	Ok its a has bean.... neither the user nor any user
	 *	charged objects exist.
	 */
	ubptr = &slot->ubh_beans;
	
	while (*ubptr != NULL) {
		if (*ubptr == ub) {
			{
			  int i;
			  for (i = 0; i < UB_RESOURCES; i++)
			    if (i != UB_RESPAGES && ub->ub_held[i])
			      printk(KERN_DEBUG "Ub %p helds %lu in %d on destroy\n",
					ub, ub->ub_held[i], i);
			}
			*ubptr = ub->ub_next;
			kmem_cache_free(ub_cachep, ub);
			unlock_beancounters(slot, flags);
			return;
		}
		ubptr = &((*ubptr)->ub_next);
	}	
#if UB_DEBUG & 1
	printk(KERN_ERR "Invalid beancounter '%p' passed to free.\n", ub);
	printk(KERN_DEBUG "Slot %p.\n", slot);
#endif
	unlock_beancounters(slot, flags);
}


/*
 *	Generic resource charging stuff
 */

static int __charge_beancounter_locked(struct user_beancounter *ub,
		int resource, unsigned long val, int strict)
{
	int retval;

#if UB_DEBUG & 2
	printk(KERN_DEBUG "Charging %lu for %d of %p with %lu\n",
			val, resource, ub, ub->ub_held[resource]);
#endif
	retval = -ENOMEM;
	/* ub_value <= UB_MAXVALUE, value <= UB_MAXVALUE, and only one addition
	 * at the moment is possible so an overflow is impossible.
	 */
	ub->ub_held[resource] += val;
	if (ub->ub_held[resource] > ub->ub_limit[resource])
		goto out;
	if (strict && ub->ub_held[resource] > ub->ub_barrier[resource])
		goto out;
	if (ub->ub_maxheld[resource] < ub->ub_held[resource])
		ub->ub_maxheld[resource] = ub->ub_held[resource];
	return 0;

out:
	ub->ub_held[resource] -= val;
	return retval;
}

int __charge_beancounter(struct user_beancounter *ub,
		int resource, unsigned long val, int strict)
{
	int retval;
	unsigned long flags;

	retval = -EINVAL;
	if (val > UB_MAXVALUE)
		goto out;

	spin_lock_irqsave(&ub->ub_lock, flags);
	retval = __charge_beancounter_locked(ub, resource, val, strict);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
out:
	return retval;
}

static inline void __uncharge_beancounter_locked(struct user_beancounter *ub,
		int resource, unsigned long val)
{
#if UB_DEBUG & 2
	printk(KERN_DEBUG "Uncharging %lu for %d of %p with %lu\n",
			val, resource, ub, ub->ub_held[resource]);
#endif
	if (ub->ub_held[resource] < val)
		printk(KERN_ERR "Uncharging %lu for %d of %p with %lu\n",
				val, resource, ub, ub->ub_held[resource]);
	ub->ub_held[resource] -= val;
}

void __uncharge_beancounter(struct user_beancounter *ub,
		int resource, unsigned long val)
{
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	__uncharge_beancounter_locked(ub, resource, val);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
}

/*
 *	Charge resources on a new task creation
 *
 *	User beancounter here is a login_bc of a parent task
 *	(which will be equal to task_bc of the new task after allocation
 *	and initialisation).
 *	Currently only stack+struct task size is charged.
 */
int __charge_task(struct user_beancounter *ub)
{
	int retval;
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	retval = __charge_beancounter_locked(ub, UB_KMEMSIZE, THREAD_SIZE, 1);
	if (!retval) {
		retval = __charge_beancounter_locked(ub, UB_NUMPROC, 1, 1);
		if (retval)
			goto fail;
	}
out:
	spin_unlock_irqrestore(&ub->ub_lock, flags);
	return retval;

fail:
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, THREAD_SIZE);
	goto out;
}

void __uncharge_task(struct user_beancounter *ub)
{
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, THREAD_SIZE);
	__uncharge_beancounter_locked(ub, UB_NUMPROC, 1);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
}

/*
 *	Different memory type accounting
 */
int __charge_memory(struct user_beancounter *ub, unsigned long size,
		unsigned vm_flags, int strict)
{
	int retval;
	unsigned long flags;

	size >>= PAGE_SHIFT;

	retval = -EINVAL;
	if (size > UB_MAXVALUE)
		goto out;

	spin_lock_irqsave(&ub->ub_lock, flags);
	retval = __charge_beancounter_locked(ub, UB_TOTVMPAGES, size, strict);
	if (retval)
		goto out_unlock;
	if (vm_flags & VM_LOCKED) {
		retval = __charge_beancounter_locked(ub, UB_LOCKEDPAGES,
				size, strict);
		if (retval)
			goto totvm_restore;
	}
	if (vm_flags & VM_ANON) {
		retval = __charge_beancounter_locked(ub, UB_ZSHMPAGES,
				size, strict);
		if (retval)
			goto locked_restore;
	}
out_unlock:
	spin_unlock_irqrestore(&ub->ub_lock, flags);
out:
	return retval;

locked_restore:
	__uncharge_beancounter_locked(ub, UB_LOCKEDPAGES, size);
totvm_restore:
	__uncharge_beancounter_locked(ub, UB_TOTVMPAGES, size);
	goto out_unlock;
}

void __uncharge_memory(struct user_beancounter *ub, unsigned long size,
		unsigned vm_flags)
{
	unsigned long flags;

	size >>= PAGE_SHIFT;

	spin_lock_irqsave(&ub->ub_lock, flags);
	__uncharge_beancounter_locked(ub, UB_TOTVMPAGES, size);
	if (vm_flags & VM_LOCKED)
		__uncharge_beancounter_locked(ub, UB_LOCKEDPAGES, size);
	if (vm_flags & VM_ANON)
		__uncharge_beancounter_locked(ub, UB_ZSHMPAGES, size);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
}


int __charge_locked_mem(struct user_beancounter *ub, unsigned long size)
{
	return __charge_beancounter(ub, UB_LOCKEDPAGES, size >> PAGE_SHIFT, 1);
}

void __uncharge_locked_mem(struct user_beancounter *ub, unsigned long size)
{
	__uncharge_beancounter(ub, UB_LOCKEDPAGES, size >> PAGE_SHIFT);
}

int __charge_kmem(struct user_beancounter *ub, unsigned long size, int strict)
{
	return __charge_beancounter(ub, UB_KMEMSIZE, size, strict);
}

void __uncharge_kmem(struct user_beancounter *ub, unsigned long size)
{
	__uncharge_beancounter(ub, UB_KMEMSIZE, size);
}

int __charge_shmpages(struct user_beancounter *ub, unsigned long size)
{
	return __charge_beancounter(ub, UB_SHMPAGES, size, 1);
}

void __uncharge_shmpages(struct user_beancounter *ub, unsigned long size)
{
	__uncharge_beancounter(ub, UB_SHMPAGES, size);
}

struct task_struct *select_worst_task(void)
{
	struct task_struct * p;
	struct task_struct * worst;
	long ub_maxover = 0;

	worst = p = init_task.next_task;
	for (; p != &init_task; p = p->next_task) {
		struct mm_struct *mm = p->mm;
		struct user_beancounter *ub;		
		long ub_overdraft = 0; /* ub current overdraft */

		if (!mm) /* don't touch the kernel threads */
			continue;
		else
 			ub = mm->beancounter;

		if (ub) {
			ub_overdraft =
				ub->ub_held[UB_TOTVMPAGES]
				- ub->ub_limit[UB_OOMGUARPAGES];
			if (ub_overdraft < 0)
				/* processes without overdraft are not
				 * preferred over ones without beancounter */
				ub_overdraft = 0;
		}
		if (ub_overdraft > ub_maxover || 
		    (ub_overdraft == ub_maxover &&
		     mm->total_vm > worst->mm->total_vm) ) {
			ub_maxover = ub_overdraft;
			worst = p;
		}
	}
	return worst;
}

/*
 *	File-related accounting
 */
int __charge_sock(struct user_beancounter *ub, unsigned long size)
{
	int retval;
	unsigned long flags;
	
	spin_lock_irqsave(&ub->ub_lock, flags);
	retval = __charge_beancounter_locked(ub, UB_KMEMSIZE, size, 1);
	if (!retval) {
		retval = __charge_beancounter_locked(ub, UB_NUMSOCK, 1, 1);
		if (retval)
			goto fail;
	}
out:
	spin_unlock_irqrestore(&ub->ub_lock, flags);
	return retval;

fail:
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, size);
	goto out;
}

void __uncharge_sock(struct user_beancounter *ub, unsigned long size)
{
	unsigned long flags;
	
	spin_lock_irqsave(&ub->ub_lock, flags);
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, size);
	__uncharge_beancounter_locked(ub, UB_NUMSOCK, 1);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
}

int __charge_flock(struct user_beancounter *ub, unsigned long size)
{
	int retval;
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	retval = __charge_beancounter_locked(ub, UB_KMEMSIZE, size, 1);
	if (!retval) {
		retval = __charge_beancounter_locked(ub, UB_NUMFLOCK, 1, 1);
		if (retval)
			goto fail;
	}
out:
	spin_unlock_irqrestore(&ub->ub_lock, flags);
	return retval;

fail:
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, size);
	goto out;
}

void __uncharge_flock(struct user_beancounter *ub, unsigned long size)
{
	unsigned long flags;
	
	spin_lock_irqsave(&ub->ub_lock, flags);
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, size);
	__uncharge_beancounter_locked(ub, UB_NUMFLOCK, 1);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
}

int __charge_pty(struct user_beancounter *ub)
{
	return __charge_beancounter(ub, UB_NUMPTY, 1, 1);
}

void __uncharge_pty(struct user_beancounter *ub)
{
	__uncharge_beancounter(ub, UB_NUMPTY, 1);
}

/*
 *	Accounting for other resources
 */
int __charge_siginfo(struct user_beancounter *ub, unsigned long size)
{
	int retval;
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	retval = __charge_beancounter_locked(ub, UB_KMEMSIZE, size, 1);
	if (!retval) {
		retval = __charge_beancounter_locked(ub, UB_NUMSIGINFO, 1, 1);
		if (retval)
			goto fail;
	}
out:
	spin_unlock_irqrestore(&ub->ub_lock, flags);
	return retval;

fail:
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, size);
	goto out;
}

void __uncharge_siginfo(struct user_beancounter *ub, unsigned long size)
{
	unsigned long flags;
	
	spin_lock_irqsave(&ub->ub_lock, flags);
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, size);
	__uncharge_beancounter_locked(ub, UB_NUMSIGINFO, 1);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
}


#else  /* defined CONFIG_USER_RESOURCE */


struct task_struct *select_worst_task(void)
{
	struct task_struct * p;
	struct task_struct * worst;

	worst = p = init_task.next_task;
	for (; p != &init_task; p = p->next_task) {
		struct mm_struct *mm = p->mm;

		if (!mm) /* don't touch the kernel threads */
			continue;

		if (mm->total_vm > worst->mm->total_vm)
			worst = p;
	}
	return worst;
}


#endif /* defined CONFIG_USER_RESOURCE */


/*
 *	Initialisation
 */
 
void __init beancounter_init(unsigned long mempages)
{
#ifdef CONFIG_USER_RESOURCE
	int k;

	ub_cachep = kmem_cache_create("user_beancounters",
			sizeof(struct user_beancounter),
			0, SLAB_HWCACHE_ALIGN,
			NULL, NULL);

	atomic_set(&default_beancounter.ub_refcount, 1);
	spin_lock_init(&default_beancounter.ub_lock);
	
	/*
	 *	Default settings
	 */
	memset(&default_beancounter.ub_held,
			0, sizeof(default_beancounter.ub_held));
	default_beancounter.ub_limit[UB_KMEMSIZE] =
				mempages > (192*1024*1024 >> PAGE_SHIFT) ?
				32*1024*1024 :
				(mempages << PAGE_SHIFT) / 6;
	default_beancounter.ub_limit[UB_LOCKEDPAGES] = 8;
	default_beancounter.ub_limit[UB_TOTVMPAGES] =
				UB_MAXVALUE; /* swappable */
	default_beancounter.ub_limit[UB_SHMPAGES] = 64;
	default_beancounter.ub_limit[UB_ZSHMPAGES] = 1024;
	default_beancounter.ub_limit[UB_NUMPROC] = max_threads / 2;
	default_beancounter.ub_limit[UB_NUMSOCK] = 1024;
	default_beancounter.ub_limit[UB_NUMFLOCK] = 1024;
	default_beancounter.ub_limit[UB_NUMPTY] = 16;
	default_beancounter.ub_limit[UB_NUMSIGINFO] = 1024;
	for (k = 0; k < UB_RESOURCES; k++)
		default_beancounter.ub_barrier[k] =
			default_beancounter.ub_limit[k];

	/*
	 *	Initialise the beancounter hash.
	 */
	for (k = 0; k < UB_HASH_SIZE; k++)
		spin_lock_init(&ub_hash[k].ubh_lock);

#ifdef CONFIG_USER_RESOURCE_PROC
	{
		struct proc_dir_entry *entry;
		extern struct file_operations ub_file_operations;
		entry = create_proc_entry("user_beancounters", S_IRUGO, NULL);
		if (entry)
			entry->proc_fops = &ub_file_operations;
	}
#endif
#endif
}


/*
 *	The (rather boring) getluid syscall
 */
asmlinkage long sys_getluid(void)
{
	return current->luid != (uid_t)-1 ? current->luid : -EINVAL;
}

/*
 *	The setluid syscall
 */
asmlinkage long sys_setluid(uid_t uid)
{
#ifdef CONFIG_USER_RESOURCE
	struct user_beancounter *ub;
#endif
	int error;
	
	/* You may not disown a setluid */
	error = -EINVAL;
	if (uid == (uid_t)-1)
		goto out;

	/* You may only set an luid as root */
	error = -EPERM;
	if (!capable(CAP_SETUID))
		goto out;

	/* The luid once set is irrevocable to all */
	if (current->luid != (uid_t)-1)
		goto out;

#ifdef CONFIG_USER_RESOURCE
	/* Ok - set up a beancounter entry for this user */
	error = -ENOBUFS;
	ub = get_beancounter_byuid(uid, 1);
	if(ub == NULL)
		goto out;
		
	printk(KERN_DEBUG "setluid, bean %p (count %d) for %.20s pid %d\n",
			ub, atomic_read(&ub->ub_refcount),
			current->comm, current->pid);
	/* Install it */
	current->login_bc = ub;
#endif	

	/* Take on our new luid and report it as OK */
	current->luid = uid;

	error = 0;
out:
	return error;
}

/*
 *	The setbeanlimit syscall
 */
asmlinkage long sys_setublimit(uid_t uid, unsigned long resource, 
		unsigned long *limits)
{
	int error;
#ifdef CONFIG_USER_RESOURCE
	unsigned long flags;
	struct user_beancounter *ub;
	unsigned long new_limits[2];

	error = -EPERM;
	if(!is_super_ve(current))                                                             
		goto out;                                                                     
	if(!capable(CAP_SYS_RESOURCE))
		goto out;

	error = -EINVAL;
	if (resource >= UB_RESOURCES)
		goto out;

	error = -EFAULT;
	if (copy_from_user(&new_limits, limits, sizeof(new_limits))) 
		goto out;

	error = -EINVAL;
	if (new_limits[0] > UB_MAXVALUE ||
	    new_limits[1] > UB_MAXVALUE)
		goto out;	

	error = -EINVAL;
	ub = get_beancounter_byuid(uid, 0);
	if (ub == NULL) { 
#if UB_DEBUG & 4
		printk(KERN_DEBUG "No login bc for uid %d\n", uid);
#endif
		goto out;
	}

	spin_lock_irqsave(&ub->ub_lock, flags);
	ub->ub_barrier[resource] = new_limits[0];
	ub->ub_limit[resource] = new_limits[1];
	spin_unlock_irqrestore(&ub->ub_lock, flags);

	__put_beancounter(ub);

	error = 0;
out:
#else
	error = -ENOSYS;
#endif
	return error;
}
