/*
 *  linux/kernel/fairsched.c
 *
 *  Fair CPU scheduler - see Documentation/fairsched.txt
 *
 *  Copyright (C) 1999-2000  Borislav Deianov <borislav@users.sourceforge.net>
 *  Copyright (C) 2000  Ensim Corporation <borislav@ensim.com>
 *
 *  Various fixes by Denis Lunev <den@asplinux.ru> and
 *                   Yury Pudgorodsky <yur@asplinux.ru>
 *
 *  Based on the QLinux scheduler by	Pawan Goyal <goyal@ensim.com>
 *					Jasleen Kaur Sahni <jks@cs.utexas.edu>
 *					T R Vishwanath <vishy@cs.utexas.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 */

#include <linux/init.h>
#include <linux/pm.h>
#include <linux/slab.h>
#include <linux/smp.h>
#include <linux/types.h> 
#include <linux/proc_fs.h>
#include <linux/vmalloc.h>

#ifdef CONFIG_FAIRSCHED		/* goes on almost till the end */

#define FAIRSCHED_VERSION "0.44"

#define DEFAULT_NODES_MAX 256
#define MAX_GEN_ID 65536U	/* max+1, at most INT_MAX+1 */
#define ID_HASH_SIZE 256	/* must be a power of 2 */
#define MAX_WEIGHT (1UL << BITS_PER_LONG/2)
/*
 * On 32-bit architectures we need to dilute the 32-bit cycle counter
 * to make sure it still fits in a long when multiplied by the maximum
 * weight. This means that start_tags overflow no faster than the
 * cycle counter (on UP).
 */
#define CYCLES_SHIFT (32 - BITS_PER_LONG/2) 

/* can be modified via /proc/sys/kernel/fairsched-nodes-max */
int fairsched_nodes_max = DEFAULT_NODES_MAX;

/* 
 * This is a pretend task_struct. Only schedule() looks at it and then
 * only at the state field, so that's all we need.
 */
static struct {
	long state;
} fake_task = { TASK_STOPPED };

#define fake_last_run ((struct task_struct *)&fake_task)

struct fairsched_node_struct fairsched_leaf;

static struct fairsched_node_struct fairsched_root = {
	id:			0,
	last_run:		fake_last_run,
	parent_node:		NULL,
	childlist_head:		LIST_HEAD_INIT(fairsched_leaf.childlist),
	runlist_head:		LIST_HEAD_INIT(fairsched_root.runlist_head),
	runqueue_head:		LIST_HEAD_INIT(fairsched_root.runqueue_head),
	nr_running:		0,
	nr_cpus:		0
};

struct fairsched_node_struct fairsched_leaf = {
	id:			1,
	weight:			1,
	start_tag:		0,
	last_run:		fake_last_run,
	parent_node:		&fairsched_root,
	childlist_head:		LIST_HEAD_INIT(fairsched_leaf.childlist_head),
	childlist:		LIST_HEAD_INIT(fairsched_root.childlist_head),
	runlist_head:		LIST_HEAD_INIT(fairsched_leaf.runlist_head),
	runqueue_head:		LIST_HEAD_INIT(fairsched_leaf.runqueue_head),
	nr_running:		0,
	nr_cpus:		0
};

static LIST_HEAD(fairsched_node_list);
static int nr_fairsched_nodes = 0;

/*
 * Locking:
 *
 * The scheduling tree data is protected by the runqueue lock. The
 * reason we don't have a separate lock is that runqueue_lock is
 * already held in all performance critical parts (all of these are
 * called from schedule() or wake_up_process[_synchronous]()).  
 */

/*
 * We align per-CPU scheduling data on cacheline boundaries,
 * to prevent cacheline ping-pong (as in sched.c)
 */
static union {
	struct schedule_data {
		unsigned long last_schedule;
	} schedule_data;
	char __pad [SMP_CACHE_BYTES];
} aligned_data [NR_CPUS] __cacheline_aligned = {{{0}}};

#define last_schedule(cpu) (aligned_data[(cpu)].schedule_data.last_schedule)

static int use_cycle_counter;

/*
 * Various support routines. Some of these are only used once but are
 * separated for clarity.
 */

static inline int has_children(struct fairsched_node_struct *node)
{
	return !list_empty(&node->childlist_head);
}

static inline int has_run_children(struct fairsched_node_struct *node)
{
	return !list_empty(&node->runlist_head);
}

static inline int in_runlist(struct fairsched_node_struct *node)
{
	return node->runlist.next != NULL;
}

static inline int has_procs(struct fairsched_node_struct *node)
{
	struct task_struct *p;

 	for_each_task_all(p)
		if (p->fairsched_node == node && p->state!=TASK_ZOMBIE)
			return 1;

	return 0;
}

static inline unsigned long ulong_max(unsigned long a, unsigned long b)
{
	return a > b ? a : b;
}

/* 
 * Return the start_tag of the first runnable child, or 0
 */
static inline unsigned long virtual_time(struct fairsched_node_struct *node)
{
	return has_run_children(node) ? 
		list_entry(node->runlist_head.next, 
			   struct fairsched_node_struct, runlist)->start_tag : 
		0;
}

/*
 * Set the start_tag of all non-runnable chidren to 0
 */
static inline void zero_nonrun_start_tags(struct fairsched_node_struct *node)
{
	struct fairsched_node_struct *node1;
	struct list_head *tmp;

	list_for_each(tmp, &node->childlist_head) {
		node1 = list_entry(tmp, struct fairsched_node_struct, 
				   childlist);
		if (!in_runlist(node1))
			node1->start_tag = 0;
	}
}

/*
 * Shift the start_tags of runnable children down to avoid overflows
 */
static inline void shift_run_start_tags(struct fairsched_node_struct *node)
{
	struct fairsched_node_struct *node1;
	struct list_head *tmp;
	unsigned long min_tag = virtual_time(node);

	list_for_each(tmp, &node->runlist_head) {
		node1 = list_entry(tmp, struct fairsched_node_struct, runlist);
		node1->start_tag -= min_tag;
	}
}

/*
 * Insert a node in a parent's runlist (sorted by start_tag)
 */
static inline void insert_run_node(struct fairsched_node_struct *node)
{
	struct fairsched_node_struct *node1;
	struct list_head *tmp;

	list_for_each(tmp, &node->parent_node->runlist_head) {
		node1 = list_entry(tmp, struct fairsched_node_struct, runlist);
		if (node->start_tag <= node1->start_tag)
			break;
	}
	/* insert node just before tmp */
	list_add_tail(&node->runlist, tmp);
}
	

/*
 * The heart of the HSFQ algorithm: fairsched_setrun, fairsched_sleep,
 * fairsched_update and fairsched_schedule 
 */

/*
 * Called when the first process is woken up in a leaf node
 */
void fairsched_setrun(struct fairsched_node_struct *node)
{
	/* 
	 * This is always called with runqueue_lock held and
	 * interrupts disabled.
	 */

	/* 
	 * Loop invariant: Something woke up in this node. If node is
	 * not already in the parent's runlist, we need to insert
	 * it. 
	 */
	while (node != &fairsched_root && !in_runlist(node)) {
		node->start_tag = ulong_max(node->start_tag,
					    virtual_time(node->parent_node));
		insert_run_node(node);

		node = node->parent_node;
	}
}

/*
 * Called when the last process is put to sleep in a leaf node
 */
void fairsched_sleep(struct fairsched_node_struct *node)
{
	/* 
	 * This is always called with runqueue_lock held and
	 * interrupts disabled...
	 *
	 * ...except on SMP boot, but those calls are serialized
	 * since only one CPU is running at that time.
	 */

	/* 
	 * Loop invariant: Something went to sleep in this node. If
	 * nothing else remains, set the start_tags of all children to
	 * 0 and delete node from parent's runlist. 
	 */
	while (node != &fairsched_root && !has_run_children(node)) {
		zero_nonrun_start_tags(node);
		list_del(&node->runlist);
		node->runlist.next = NULL;

		node = node->parent_node;
	}
}

/* from sched.c */
#ifdef CONFIG_SMP
#define idle_task(cpu) (init_tasks[cpu_number_map(cpu)])
#else
#define idle_task(cpu) (&init_task)
#endif

/*
 * Called when a process relinquishes a CPU
 *
 * When a cycle counter is available, we can use that to calculate
 * duration. Otherwise, we'll have to make do with jiffies. Note that
 * jiffies*MHz would have been a bit better if we have processors with
 * different speeds but we don't have a reliable way of determining
 * the clock frequency in the absence of a cycle counter.
 */
void fairsched_update(struct task_struct *last_run, int this_cpu)
{
	struct fairsched_node_struct *node = last_run->fairsched_node;
	unsigned long duration;

	/* 
	 * This is always called with runqueue_lock held and
	 * interrupts disabled.
	 */

	if (use_cycle_counter) {
		cycles_t t = get_cycles();
		/* round up, only least valuable part of long */
		duration = (t - (cycles_t)last_schedule(this_cpu)) +
			(1 << CYCLES_SHIFT) - 1;
		duration >>= CYCLES_SHIFT;
		last_schedule(this_cpu) = t;
	} else {
		/* round up */
		duration = jiffies - last_schedule(this_cpu) + 1;
		last_schedule(this_cpu) = jiffies;
	}
	node->used_time += duration;

	/*
	 * Don't store an idle task in last_run. It is pointless
	 * (schedule() already defaults to an idle task if no other
	 * process is runnable) and dangerous (that node may run next
	 * on another CPU which cannot run this idle task).
	 *
	 * Also, zombies can never run again.
	 *
	 * BTW, to get as close as possible to the original scheduler
	 * semantics in each node we may need per-processor last_run
	 * (for example to get correct scheduling order for FIFO
	 * tasks).  On the other hand, last_run handling was very hard
	 * to get right, so I don't feel like touching it at all...  
	 */
	if (last_run == idle_task(this_cpu) || last_run->state == TASK_ZOMBIE)
		node->last_run = fake_last_run;
	else
		node->last_run = last_run;

	/* Avoid charging the idle task to fairsched_leaf */
	if (last_run == idle_task(this_cpu))
		goto out;

	fairsched_update_nr_cpus(node, -1);

	/* Loop invariant: in_runlist(node) */
	while (node != &fairsched_root) {
		unsigned long inc = duration * node->weight;
		struct fairsched_node_struct *parent = node->parent_node;

		/* ulong overflow protection */
		if (node->start_tag + inc < node->start_tag) {
			zero_nonrun_start_tags(parent);
			shift_run_start_tags(parent);
		}

		node->start_tag += inc;
		list_del(&node->runlist);
		insert_run_node(node);

		node = parent;
	}

 out:
}
		
/*
 * Chooses the next node to run a process from
 */
struct fairsched_node_struct *fairsched_schedule(void)
{
	struct fairsched_node_struct *node;

	/* 
	 * This is always called with runqueue_lock held and
	 * interrupts disabled.
	 */

#ifdef CONFIG_SMP
	if (fairsched_root.nr_running > fairsched_root.nr_cpus) {
		/*
		 * There are some runnable tasks that are not running
		 */

		node = &fairsched_root;
		/*
		 * Invariant: If node has runnable children then at
		 * least one of them is hungry (contains more runnable
		 * tasks than the number of CPUs it occupies).  
		 */
		while (has_run_children(node)) {
			/* go down to first runnable child */
			node = list_entry(node->runlist_head.next, 
					  struct fairsched_node_struct, 
					  runlist);
			while (node->nr_running == node->nr_cpus)
				/* go sideways to first hungry sibling */
				node = list_entry(node->runlist.next,
						  struct fairsched_node_struct,
						  runlist);
		}

		fairsched_update_nr_cpus(node, +1);
	} else {
		/*
		 * All runnable tasks already have CPUs, nothing left
		 * for us.
		 */
		node = NULL;
	}
#else

	node = &fairsched_root;
	while(has_run_children(node))
		node = list_entry(node->runlist_head.next, 
				  struct fairsched_node_struct, runlist);

	/* No runnable nodes */
	if (node == &fairsched_root)
		node = NULL;
#endif

	return node;
}


/* 
 * Hashing for node IDs 
 */
static struct list_head id_hash[ID_HASH_SIZE];

#define id_hashfn(x) ((x) & (ID_HASH_SIZE - 1))

static inline void hash_node(struct fairsched_node_struct *node)
{
	list_add(&node->hashlist, &id_hash[id_hashfn(node->id)]);
	list_add_tail(&node->nodelist, &fairsched_node_list);
	nr_fairsched_nodes++;
}

static inline void unhash_node(struct fairsched_node_struct *node)
{
	nr_fairsched_nodes--;
	list_del(&node->nodelist);
	list_del(&node->hashlist);
}

static inline struct fairsched_node_struct *find_node_by_id(unsigned int id)
{
	struct fairsched_node_struct *node;
	struct list_head *tmp;

	list_for_each(tmp, &id_hash[id_hashfn(id)]) {
		node = list_entry(tmp, struct fairsched_node_struct, hashlist);
		if (node->id == id && check_ve_strict(node,current))
			return node;
	}

	return NULL;
}

static inline unsigned int get_id(unsigned int desired)
{
	static unsigned int last_id = 1;
	static unsigned int next_safe = MAX_GEN_ID;
	struct fairsched_node_struct *node;
	struct list_head *tmp;
	unsigned int retval;

	if (desired != 0) {
		if (desired > last_id && next_safe > desired)
			next_safe = desired;
		retval = desired;
		goto out;
	}

	if (++last_id == MAX_GEN_ID) {
		last_id = 2;
		goto inside;
	}

	if (last_id >= next_safe) {

	inside:
		next_safe = MAX_GEN_ID;
	repeat:
		list_for_each(tmp, &fairsched_node_list) {
			node = list_entry(tmp, struct fairsched_node_struct, 
					  nodelist);
			if (node->id == last_id) {
				if (++last_id >= next_safe) {
					if (last_id >= MAX_GEN_ID)
						last_id = 2;
					next_safe = MAX_GEN_ID;
				}
				goto repeat;
			}
			if (node->id > last_id && next_safe > node->id)
				next_safe = node->id;
		}
	}

	retval = last_id;

	if (retval >= MAX_GEN_ID)
		BUG();

 out:
	if (find_node_by_id(retval) != NULL)
		BUG();

	return retval;
}
	

/* 
 * System calls 
 */

asmlinkage long sys_fairsched_mknod(unsigned int parent, unsigned int wght,
				    unsigned int desired)
{
	unsigned long weight = (unsigned long)wght;
	struct fairsched_node_struct *parent_node, *new_node;
	unsigned long flags;
	long retval;

	retval = -EINVAL;
	if (weight < 1 || weight >= MAX_WEIGHT)
		goto out;

	new_node = kmalloc(sizeof(*new_node), GFP_KERNEL);

	spin_lock_irqsave(&runqueue_lock, flags);
	read_lock(&tasklist_lock);

	retval = -EINVAL;
	parent_node = find_node_by_id(parent);
	if (parent_node == NULL)
		goto unlock;

	retval = -EBUSY;
	if (has_procs(parent_node))
		goto unlock;
	if (desired != 0 && find_node_by_id(desired) != NULL)
		goto unlock;

	retval = -ENOMEM;
	if (nr_fairsched_nodes >= fairsched_nodes_max)
		goto unlock;
	if (new_node == NULL)
		goto unlock;

	retval = -EPERM;
	if (!capable(CAP_SYS_NICE))
		goto unlock;

	memset( new_node, 0, sizeof(struct fairsched_node_struct) );
	new_node->id = get_id(desired);
	new_node->weight = weight;
	new_node->parent_node = parent_node;
	INIT_LIST_HEAD(&new_node->childlist_head);
	list_add(&new_node->childlist, &parent_node->childlist_head);
	INIT_LIST_HEAD(&new_node->runlist_head);
	INIT_LIST_HEAD(&new_node->runqueue_head);
	new_node->last_run = fake_last_run;
#ifdef CONFIG_VE
	new_node->envid = current->envid;
#endif

	hash_node(new_node);

	/* desired might be greater than LONG_MAX */
	retval = desired == 0 ? new_node->id : 0;

 unlock:
	read_unlock(&tasklist_lock);
	spin_unlock_irqrestore(&runqueue_lock, flags);

	if (retval < 0 && new_node != NULL)
		kfree(new_node);
 out:
	return retval;
}

asmlinkage long sys_fairsched_rmnod(unsigned int id)
{
	struct fairsched_node_struct *node;
	unsigned long flags;
	long retval;

	spin_lock_irqsave(&runqueue_lock, flags);
	read_lock(&tasklist_lock);

	retval = -EINVAL;
	node = find_node_by_id(id);
	if (!node)
		goto unlock;

	retval = -EBUSY;
	if (has_children(node) || has_procs(node))
		goto unlock;

	retval = -EPERM;
	if (!capable(CAP_SYS_NICE))
		goto unlock;

	list_del(&node->childlist);
	unhash_node(node);
	kfree(node);

	retval = 0;

 unlock:
	read_unlock(&tasklist_lock);
	spin_unlock_irqrestore(&runqueue_lock, flags);

	return retval;
}

asmlinkage long sys_fairsched_chwt(unsigned int id, unsigned int wght)
{
	unsigned long weight = (unsigned long)wght;
	struct fairsched_node_struct *node;
	unsigned long flags;
	long retval;

	retval = -EINVAL;
	if (weight < 1 || weight >= MAX_WEIGHT || id == 0)
		goto out;

	spin_lock_irqsave(&runqueue_lock, flags);

	retval = -EINVAL;
	node = find_node_by_id(id);
	if (node == NULL)
		goto unlock;

	retval = -EPERM;
	if (!capable(CAP_SYS_NICE))
		goto unlock;

	node->weight = weight;
	retval = 0;

 unlock:
	spin_unlock_irqrestore(&runqueue_lock, flags);
 out:
	return retval;
}

/* 
 * Only sleeping processes can be moved, except the current process
 * can move itself.  
 */
static long do_fairsched_mvpr(struct task_struct *p, unsigned int newid)
{
	struct fairsched_node_struct *node = find_node_by_id(newid);
	if (node == NULL)
		return -EINVAL;

	if (has_children(node)) 
		return -EBUSY;

	if ((task_on_runqueue(p) && p != current) || p->state==TASK_ZOMBIE)
		return -EBUSY;

	if (!capable(CAP_SYS_NICE))
		return -EPERM;

	if (p == current) {
		fairsched_update(p, p->processor);
		del_from_runqueue(p);
	}

	if (p->fairsched_node->last_run == p)
		p->fairsched_node->last_run = fake_last_run;

	p->fairsched_node = node;

	if (p == current) {
		add_to_runqueue(p);
		fairsched_update_nr_cpus(node, +1);
	}

	return 0;
}

asmlinkage long sys_fairsched_mvpr(pid_t pid, unsigned int newid)
{
	unsigned long flags;
	long retval;
	struct task_struct *p;

	spin_lock_irqsave(&runqueue_lock, flags);
	read_lock(&tasklist_lock);

	retval = -ESRCH;
	p = find_task_by_pid_ve(pid);
	if (p == NULL)
		goto unlock;
	
	retval = do_fairsched_mvpr(p,newid);

 unlock:
	read_unlock(&tasklist_lock);
	spin_unlock_irqrestore(&runqueue_lock, flags);

	return retval;
}


/*
 * Snapshots of the scheduling tree
 */
struct snap_node_struct {
	unsigned int id;
	unsigned int parent_id;
	unsigned int wght;
	pid_t last_pid;
	int nr_tasks;
	int nr_running;
	int nr_cpus;
	unsigned long used_time;
#ifdef CONFIG_VE
	envid_t veid;
#endif
};

static struct snap_node_struct *snap_node = NULL;
static int nr_snap_nodes;

static rwlock_t snap_lock = RW_LOCK_UNLOCKED;

static int take_snapshot(void)
{
	struct snap_node_struct *local_snap_node;
	int local_nr_snap_nodes;
	struct fairsched_node_struct *node;
	struct list_head *tmp;
	struct snap_node_struct *snap;
	struct task_struct *p;
	unsigned long flags;

 repeat:
	/* some slack in case nr_fairsched_nodes increases */
	local_nr_snap_nodes = nr_fairsched_nodes + 10;
	local_snap_node = vmalloc(local_nr_snap_nodes * 
				  sizeof(struct snap_node_struct));
	if (local_snap_node == NULL)
		return -ENOMEM;

	spin_lock_irqsave(&runqueue_lock, flags);

	if (local_nr_snap_nodes < nr_fairsched_nodes) {
		spin_unlock_irqrestore(&runqueue_lock, flags);
		vfree(local_snap_node);
		goto repeat;
	}
#ifdef CONFIG_VE
	local_nr_snap_nodes = 0;
#else
	local_nr_snap_nodes = nr_fairsched_nodes;
#endif

	list_for_each(tmp, &fairsched_node_list) {
		node = list_entry(tmp, struct fairsched_node_struct, nodelist);
		node->nr_tasks = 0;
	}
	read_lock(&tasklist_lock);
	for_each_task_all(p)
		fairsched_update_nr_tasks(p->fairsched_node, +1);
	read_unlock(&tasklist_lock);

	snap = local_snap_node;
	list_for_each(tmp, &fairsched_node_list) {
		node = list_entry(tmp, struct fairsched_node_struct, nodelist);

#ifdef CONFIG_VE
		if( !check_current_ve( node ))
			continue;
		++local_nr_snap_nodes;
		snap->veid = node->envid->veid;
#endif

		snap->id = node->id;
		snap->parent_id = node->parent_node != NULL ? 
			node->parent_node->id : 0;
		snap->wght = (unsigned int)node->weight;
		snap->last_pid = node->last_run != fake_last_run ?
			node->last_run->pid : 0;
		snap->nr_tasks = node->nr_tasks;
		snap->nr_running = node->nr_running;
		snap->nr_cpus = node->nr_cpus;
		snap->used_time = node->used_time;

		snap++;
	}

	spin_unlock_irqrestore(&runqueue_lock, flags);

	/* install the new snapshot atomically */
	write_lock(&snap_lock);

	vfree(snap_node);
	snap_node = local_snap_node;
	nr_snap_nodes = local_nr_snap_nodes;

	write_unlock(&snap_lock);

	return 0;
}

static int fairsched_read_proc(char *page, char **start, off_t off, int count, 
			  int *eof, void *data)
{
	struct snap_node_struct *snap;
	int len = 0;
	int old_len = 0;

	if (off == 0 || snap_node == NULL)
		if ((len = take_snapshot()) < 0)
			goto out;

	read_lock(&snap_lock);

	snap = snap_node + off;
	while (snap < snap_node + nr_snap_nodes && len < count) {
		old_len = len;
		if (snap == snap_node)
			len += sprintf(page + len,
#ifdef CONFIG_VE
		"      veid"
#endif
		"         id     parent     weight   pid tasks running cpus used_time\n");

		len += sprintf(page + len,
#ifdef CONFIG_VE
			       "%10u %10u %10u %10u %5d %5d %7d %4d %10lu\n",
			       snap->veid,
#else
			       "%10u %10u %10u %5d %5d %7d %4d %10lu\n",
#endif
			       snap->id,
			       snap->parent_id,
			       snap->wght,
			       snap->last_pid,
			       snap->nr_tasks,
			       snap->nr_running,
			       snap->nr_cpus,
			       snap->used_time);
		snap++;
	}

	if (len > count) {
		len = old_len;
		snap--;
	}
	*(long *)start = snap - (snap_node + off);	/* yuk */

	read_unlock(&snap_lock);
 out:
	return len;
}


#ifdef CONFIG_PM
/*
 * After a resume the cycle counter may be reset. This is bad because
 * the next duration calculation will be wildly wrong and may lead to
 * some nodes being locked out for a loooong time.
 *
 * We zero all start_tags and last_schedule() on a resume event. This
 * means that until the callback is called scheduling will not be
 * fair, but after that everything is reset and back to square 0.
 */
static int pm_call(struct pm_dev *dev, pm_request_t rqst, void *data)
{
	unsigned long flags;
	struct fairsched_node_struct *node;
	struct list_head *tmp;
	int i;

	if (rqst == PM_RESUME) {
		spin_lock_irqsave(&runqueue_lock, flags);

		list_for_each(tmp, &fairsched_node_list) {
			node = list_entry(tmp, struct fairsched_node_struct, 
					  nodelist);
			node->start_tag = 0;
		}

		for (i=0; i<NR_CPUS; i++)
			last_schedule(i) = 0;
		
		spin_unlock_irqrestore(&runqueue_lock, flags);
	}

	return 0;
}
#endif


static int __init fairsched_init(void)
{
	int i;

	printk("Fair CPU scheduler v%s\n", FAIRSCHED_VERSION);

	for (i=0; i<ID_HASH_SIZE; i++)
		INIT_LIST_HEAD(&id_hash[i]);
	hash_node(&fairsched_root);
	hash_node(&fairsched_leaf);

	use_cycle_counter = (get_cycles() != 0);

#ifdef CONFIG_PM
	if (use_cycle_counter)
		(void)pm_register(PM_SYS_DEV, PM_SYS_UNKNOWN, pm_call);
#endif

	create_proc_read_entry("fairsched", 0, NULL, &fairsched_read_proc, 
			       NULL);

#ifdef CONFIG_VE
	fairsched_root.envid = fairsched_leaf.envid = child_reaper->envid;
#endif

	return 0;
}

module_init(fairsched_init);

#ifdef CONFIG_VE
int ve_fairsched_attach( struct ve_struct *envid )
{
	long err;
	unsigned hidden_root = 0, root = 0, child = 0;
	struct fairsched_node_struct *node;
	unsigned long flags;

	if( (err = sys_fairsched_mknod( 0, 1, 0 ))<0 )
		goto out;
	hidden_root = (unsigned)err;

	if( (err = sys_fairsched_mknod( hidden_root, 1, 0 ))<0 )
		goto out;
	root = (unsigned)err;

	if( (err = sys_fairsched_mknod( root, 1, 0 ))<0 )
		goto out;
	child = (unsigned)err;

	spin_lock_irqsave(&runqueue_lock, flags);
	read_lock( &tasklist_lock );

	if( (err = do_fairsched_mvpr( current, child ))<0 )
		goto out_unlock;

	node = find_node_by_id( child );
	if( !node )
		BUG();
	
	node->envid = node->parent_node->envid = envid;
	node->id = 1;
	node->parent_node->id = 0;
	
	read_unlock( &tasklist_lock );
	spin_unlock_irqrestore(&runqueue_lock, flags);
	return 0;

  out_unlock:
	read_unlock( &tasklist_lock );
	spin_unlock_irqrestore(&runqueue_lock, flags);
  out:
	sys_fairsched_rmnod( child );
	sys_fairsched_rmnod( root );
	sys_fairsched_rmnod( hidden_root );
	return err;
}

void ve_fairsched_detach( struct ve_struct *envid )
{
	struct fairsched_node_struct *parent = NULL, *node;
	unsigned long flags;
	int counter;

	spin_lock_irqsave(&runqueue_lock, flags);
	read_lock( &tasklist_lock );

	do
	{
		struct list_head *tmp;
		
		counter = 0;
		list_for_each(tmp, &fairsched_node_list) {
			node = list_entry(tmp, struct fairsched_node_struct, nodelist);
			if( node->envid!=envid )
				continue;
			counter++;
			if (has_children(node) || has_procs(node))
				continue;

			parent = node->parent_node;
			list_del(&node->childlist);
			unhash_node(node);
			kfree(node);
		}
	} while( counter );
	
	if( parent && !has_children(parent) && !has_procs(parent) )
	{
		list_del(&parent->childlist);
		unhash_node(parent);
		kfree(parent);
	}
	read_unlock( &tasklist_lock );
	spin_unlock_irqrestore(&runqueue_lock, flags);
}

void __init fairsched_ve_init(struct ve_struct *envid)
{
	struct list_head *tmp;
	struct fairsched_node_struct *node;
	unsigned long flags;
	
	spin_lock_irqsave(&runqueue_lock, flags);
	read_lock( &tasklist_lock );

	list_for_each(tmp, &fairsched_node_list) {
		node = list_entry(tmp, struct fairsched_node_struct, nodelist);
		node->envid = envid;
	}
	
	read_unlock( &tasklist_lock );
	spin_unlock_irqrestore(&runqueue_lock, flags);
}
#endif

#else /* CONFIG_FAIRSCHED */


/* dummy syscalls if we are not using the Fair CPU scheduler */

asmlinkage long sys_fairsched_mknod(unsigned int parent, unsigned int wght,
				    unsigned int desired)
{
	return -ENOSYS;
}

asmlinkage long sys_fairsched_rmnod(unsigned int id)
{
	return -ENOSYS;
}

asmlinkage long sys_fairsched_chwt(unsigned int id, unsigned int wght)
{
	return -ENOSYS;
}

asmlinkage long sys_fairsched_mvpr(pid_t pid, unsigned int newid)
{
	return -ENOSYS;
}

#endif /* CONFIG_FAIRSCHED */
