/*
 *  linux/kernel/ve.c
 *
 *  Copyright (C) 2000  SWSoft Pte Ltd
 *
 *  Author Denis V. Lunev den@asp-linux.com
 *
 *  2000-03-10  Created
 */

/*
 * 've.c' is file with basic VE support. It provides basic primities
 * along with initialization script
 */

#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/capability.h>
#include <linux/ve.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/list.h>

#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/malloc.h>
#include <linux/sys.h>
#include <linux/kdev_t.h>
#include <linux/tty_driver.h>
#include <linux/netdevice.h>
#include <linux/utsname.h>
#include <linux/proc_fs.h>
#include <linux/kernel_stat.h>

/* Various externals from other parts of kernel */
extern struct task_struct *child_reaper;
extern struct tty_driver pty_driver, pty_slave_driver;
extern struct tty_driver *tty_drivers;

#ifdef CONFIG_UNIX98_PTYS
extern struct tty_driver ptm_driver[];	/* Unix98 pty masters; for /dev/ptmx */
extern struct tty_driver pts_driver[];	/* Unix98 pty slaves;  for /dev/ptmx */
#endif

extern rwlock_t  tty_driver_guard;
extern int get_pid(unsigned long flags);

#ifdef CONFIG_SYSVIPC
extern int ve_ipc_cleanup( struct ve_struct *envid );
#endif

#ifdef CONFIG_VE_LINK
static struct sock *nlve0 = NULL;
#endif

static struct ve_struct *ve_list_head = NULL;
rwlock_t ve_list_guard = RW_LOCK_UNLOCKED;
rwlock_t devperms_hash_guard = RW_LOCK_UNLOCKED;

asmlinkage int  sys_env_create(envid_t veid, unsigned flags, u32 addr);
asmlinkage int  sys_mark_env_to_down(envid_t veid);
static int 	do_env_create(envid_t veid, unsigned flags, u32 addr);
static int	do_prepare_devperms( envid_t veid, unsigned mode );
static void	do_clean_devperms( envid_t veid );
static int	register_ve_tty_driver( struct ve_struct* envid );
static void	unregister_ve_tty_driver( struct ve_struct* envid );
static void	clear_termios( struct tty_driver* driver );
static int	ve_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data);
static int	ve_get_info(char *buffer, off_t offset, int length);

/* Fair scheduler */
#ifdef CONFIG_FAIRSCHED
int ve_fairsched_attach( struct ve_struct *envid );
int ve_fairsched_detach( struct ve_struct *envid );
void __init fairsched_ve_init(struct ve_struct *envid);
#endif

#ifdef CONFIG_DEVPTS_FS
int ve_register_devpts_fs( struct ve_struct *envid );
void ve_unregister_devpts_fs( struct ve_struct *envid );
int __init ve_devpts_fs_init( struct ve_struct *envid );
#endif


/* Default capabilities list for VE */
kernel_cap_t  env_cap_default;


static inline struct ve_struct *get_ve(envid_t veid)
{
	struct ve_struct *ptr;
	read_lock( &ve_list_guard );
	for( ptr = ve_list_head; ptr; ptr = ptr->next )
		if( ptr->veid == veid )
			break;
	read_unlock( &ve_list_guard );
	return ptr;
}

/* Devices permissions routines,
 * character and block devices separately
 */
/* Rules applied in the following order:
   MAJOR!=0, MINOR!=0
   MAJOR!=0, MINOR==0
   MAJOR==0, MINOR==0
*/
struct devperms_struct
{
	kdev_t   	dev;	/* device id */
	unsigned 	type;
	envid_t	 	veid;
	unsigned char	mask;	

	struct devperms_struct *devhash_next;
	struct devperms_struct **devhash_pprev;
};

static struct devperms_struct original_perms[] =
{{
	MKDEV(0,0),	/*device*/
	S_IFCHR,	/*type*/
	0,		/*veid*/
	S_IRWXO,
	NULL, NULL
},
{
	MKDEV(0,0),	/*device*/
	S_IFBLK,	/*type*/
	0,		/*veid*/
	S_IRWXO,
	NULL, NULL
}};
	
static unsigned default_chrperms[] = {
	PTY_MASTER_MAJOR, 0, S_IRWXO,
	PTY_SLAVE_MAJOR, 0, S_IRWXO,
	TTY_MAJOR, 0, S_IRWXO,
	TTYAUX_MAJOR, 0, S_IRWXO,
	UNIX98_PTY_MASTER_MAJOR, 0, S_IRWXO,
	UNIX98_PTY_SLAVE_MAJOR, 0, S_IRWXO,
	MEM_MAJOR, 5, S_IRWXO, 			/* zero */
	MEM_MAJOR, 8, S_IRWXO, 			/* random */
	MEM_MAJOR, 9, S_IRWXO, 			/* urandom */
	MEM_MAJOR, 3, S_IRWXO, 			/* null */
	0
};
static unsigned default_blkperms[] = {
	0
};


#define DEVPERMS_HASH_SZ 64
struct devperms_struct *devperms_hash[DEVPERMS_HASH_SZ];

#define devperms_hashfn(id,dev) ( id<<8 ^ (MAJOR(dev)<<3) ^ MINOR(dev) ) & (DEVPERMS_HASH_SZ - 1)

static inline void hash_devperms(struct devperms_struct *p)
{
	struct devperms_struct **htable = &devperms_hash[devperms_hashfn(p->veid,p->dev)];

	if((p->devhash_next = *htable) != NULL)
		(*htable)->devhash_pprev = &p->devhash_next;
	*htable = p;
	p->devhash_pprev = htable;
}

static inline void unhash_devperms(struct devperms_struct *p)
{
	if(p->devhash_next)
		p->devhash_next->devhash_pprev = p->devhash_pprev;
	*p->devhash_pprev = p->devhash_next;
}

static inline struct devperms_struct *find_devperms(envid_t veid, int type, kdev_t dev)
{
	struct devperms_struct *p, **htable = &devperms_hash[devperms_hashfn(veid,dev)];
	
	for(p = *htable; p && !( p->type==(type&S_IFMT) &&
				 MAJOR(dev)==MAJOR(p->dev) &&
				 MINOR(dev)==MINOR(p->dev) &&
				 p->veid==veid);
	    p = p->devhash_next )
		;
	return p;
}

static int do_prepare_devperms( envid_t veid, unsigned mode )
{
	unsigned *ptr = S_ISCHR(mode) ? default_chrperms : default_blkperms;
	struct devperms_struct *new = kmalloc( sizeof(struct devperms_struct), GFP_KERNEL );

	if( !new )
		return -ENOMEM;

	new->veid = veid;
	new->type = mode;
	new->dev = MKDEV(0,0);
	new->mask = 0;
	hash_devperms( new );

	while( *ptr )
	{
		unsigned major = *ptr++;
		unsigned minor = *ptr++;
		new = kmalloc( sizeof(struct devperms_struct), GFP_KERNEL );
		if( !new )
			return -ENOMEM;
		
		new->veid = veid;
		new->type = mode;
		new->dev = MKDEV(major,minor);
		new->mask = *ptr++;
		hash_devperms( new );
	}
	return 0;
}

static void do_clean_devperms( envid_t veid ) 
{
	int i;
	struct devperms_struct* ptr;

	for( i=0; i<DEVPERMS_HASH_SZ; i++ )
		for( ptr = devperms_hash[i]; ptr; )
		{
			struct devperms_struct *next = ptr->devhash_next;
			if( ptr->veid==veid )
			{
				unhash_devperms( ptr );
				kfree( ptr );
			}		
				
			ptr = next;
		}
}

unsigned char get_device_perms_ve( envid_t veid, int type, kdev_t dev )
{
	struct devperms_struct *ptr;

	read_lock( &devperms_hash_guard );
	ptr = find_devperms( veid, type, dev );

	if( ptr )
		goto end;
	if( MINOR(dev) )
		ptr = find_devperms( veid, type, MKDEV(MAJOR(dev),0) );
	if( ptr )
		goto end;

	if( MAJOR(dev) )
		ptr = find_devperms( veid, type, MKDEV(0,0) );
	
end:	read_unlock( &devperms_hash_guard );
	return ptr ? ptr->mask : 0;
}

asmlinkage int sys_setdevperms( envid_t veid, unsigned type, kdev_t dev, unsigned mask ) 
{
	struct ve_struct   	 *ptr;
	struct devperms_struct   *perms;
	
	if( !capable(CAP_SETVEID) || veid==0 )
		return -EPERM;

	if( S_ISBLK(type) && MAJOR(dev)>=MAX_BLKDEV )
		return -ENODEV;

	if( S_ISCHR(type) && MAJOR(dev)>=MAX_CHRDEV )
		return -ENODEV;

	if( (ptr=get_ve(veid))==NULL )
		return -ESRCH;

	write_lock_irq( &devperms_hash_guard );
	perms = find_devperms( veid, type, dev );
	if( !perms )
	{
		perms = kmalloc( sizeof(struct devperms_struct), GFP_KERNEL );
		if( !perms ) {
			write_unlock_irq( &devperms_hash_guard );
			return -ENOMEM;
		}
			
		perms->veid = veid;
		perms->dev = dev;
		perms->type = type;
		perms->mask = mask & S_IRWXO;
		hash_devperms( perms );
	} else
		perms->mask = mask & S_IRWXO;
	write_unlock_irq( &devperms_hash_guard );

	return 0;
}


static int ve_read_proc(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	int len = ve_get_info(page, off, count);
	*start = page;
	if (len>count) len = count;
	if (len<0) len = 0;
	*eof = len < count;
	page[len]=0;
	return len;
}


static int ve_get_info(char *page, off_t offset, int length)
{
	off_t len = 0, clen, out_count = 0;
	char buffer[128];
	int rsize = 0;
	struct ve_struct *ptr;

/*	if( !is_super_ve(current) )
		return sprintf( page, "%10d %5d\n", current->envid->veid,
		atomic_read(&current->envid->pcounter) );*/
	
	*page = 0;
	read_lock( &ve_list_guard );
	for( ptr = is_super_ve(current) ? ve_list_head : get_current_veid();
	     ptr; ptr = is_super_ve(current) ? ptr->next : NULL ) {
		int pos;
		out_count += rsize;
		if( out_count<offset && rsize )
			continue;
		clen = sprintf( buffer, "%10d %5d\n", ptr->veid,
				atomic_read(&ptr->pcounter) );
		if( !rsize )
		{
			out_count = rsize = clen;
			if( out_count < offset )
				continue;
		}

		pos = len+offset-out_count+rsize;
		strncat( page+len, buffer+pos, length-len );
		pos = strlen(buffer+pos);
		len += pos<length-len ? pos : length-len;
		if( len>length )
			break;
	}
	read_unlock( &ve_list_guard );
	return len;
}


asmlinkage int  sys_env_create(envid_t veid, unsigned flags, u32 addr)
{
	int status = 0;

	if( !flags )
		return current->envid->veid;
	
	if( !capable(CAP_SETVEID) ) 
		return -EPERM;

	if( flags&VE_CREATE )
		flags |= VE_ENTER;
	if( flags&VE_TEST && flags&VE_ENTER )
		return -EINVAL;

	lock_kernel();
	status = do_env_create(veid, flags, addr);
	unlock_kernel();
	return status;
}

asmlinkage int  sys_mark_env_to_down(envid_t veid)
{
	struct ve_struct *ptr;
	
	if( !capable(CAP_SETVEID) || veid==0 ) 
		return -EPERM;

	if( (ptr=get_ve(veid))==NULL )
		return -ESRCH;

	do_mark_env_to_down(ptr);
	return 0;
}

void do_mark_env_to_down( struct ve_struct *ptr )
{
	int pid = get_pid(0);

	write_lock_irq( &tasklist_lock );
	unhash_pid( ptr->init_entry );
	ptr->init_entry->pid = pid;
	hash_pid( ptr->init_entry );
	write_unlock_irq( &tasklist_lock );
}

static int do_env_create(envid_t veid, unsigned flags, u32 addr)
{
	struct task_struct *proc = current;
	struct ve_struct *old = proc->envid;
	struct ve_struct *ptr = get_ve(veid);
	int    err = 0;

	if( ptr )
	{
		if( flags&VE_EXCLUSIVE )
			return -EACCES;
		else
			goto out_success;
	}
	
	if( !(flags & VE_CREATE) )
		return -ESRCH;

	ptr = kmalloc( sizeof(struct ve_struct), GFP_KERNEL );
	if( ptr == NULL )
		return -ENOMEM;

	memset( ptr, 0, sizeof(struct ve_struct) );
	memcpy( &ptr->_system_utsname, &child_reaper->envid->_system_utsname,
		sizeof(struct new_utsname) );
	ptr->veid = veid;
	memcpy( &ptr->cap_default, &env_cap_default, sizeof(env_cap_default) );
	ptr->ip = addr;
	sema_init( &ptr->init_exit_guard, 1 );
	
	if( (err = register_ve_tty_driver(ptr))!=0 )
		goto err_vetty;

	write_lock_irq( &devperms_hash_guard );
	if( (err=do_prepare_devperms( veid, S_IFCHR ))!=0 )
		goto err_perms;
	if( (err=do_prepare_devperms( veid, S_IFBLK ))!=0 )
		goto err_perms;
	write_unlock_irq( &devperms_hash_guard );

	ptr->init_entry = proc;

	/* 
	 * Save cpu stats for consitent uptime/cpu stat output:
	 *    Stubborn Linus does not export jiffies to userland,
	 *    so userland compute it from tricky uptime <-> cpu ticks relation.
	 *    Very, very tricky...
	 */
	ptr->per_cpu_user[NR_CPUS] = kstat.per_cpu_user[NR_CPUS];
	ptr->per_cpu_nice[NR_CPUS] = kstat.per_cpu_nice[NR_CPUS];
	ptr->per_cpu_system[NR_CPUS] = kstat.per_cpu_system[NR_CPUS];
	ptr->idle = init_tasks[0]->times.tms_utime + init_tasks[0]->times.tms_stime;

#ifdef CONFIG_FAIRSCHED
	ve_fairsched_attach( ptr );
#endif
	
	read_lock(&proc->fs->lock);
	ptr->fs_rootmnt = proc->fs->rootmnt;
	ptr->fs_root = proc->fs->root;
	read_unlock(&proc->fs->lock);
	
	write_lock_irq( &tasklist_lock );
	unhash_pid(proc);
	proc->pid = 1;
	hash_pid(proc);	
	write_unlock_irq( &tasklist_lock );

	write_lock_irq( &ve_list_guard );
	ptr->prev = NULL;
	ptr->next = ve_list_head;
	if( ve_list_head )
		ve_list_head->prev = ptr;
	ve_list_head = ptr;
	write_unlock_irq( &ve_list_guard );

  out_success:
	if( flags&VE_ENTER )
	{
		if( down_trylock( &ptr->init_exit_guard ))
			return -EBUSY;
		up( &ptr->init_exit_guard );

		cap_mask( proc->cap_effective, ptr->cap_default );
		cap_mask( proc->cap_inheritable, ptr->cap_default );
		cap_mask( proc->cap_permitted, ptr->cap_default );
		proc->envid = ptr;
		atomic_inc( &ptr->pcounter );
	}
	return proc->envid->veid;

  err_perms:
	do_clean_devperms( ptr->veid );
	write_unlock_irq( &devperms_hash_guard );
  err_vetty:
	unregister_ve_tty_driver(ptr);
	kfree(ptr);
	proc->envid = old;
	return err;
}

void do_env_cleanup( struct ve_struct *ptr )
{
	struct ve_struct *envid;
	lock_kernel();
	envid = current->envid;
	current->envid = ptr;
	
#ifdef CONFIG_SYSVIPC
	ve_ipc_cleanup( ptr );
#endif
	unregister_ve_tty_driver( ptr );
	
	write_lock_irq( &devperms_hash_guard );
	do_clean_devperms( ptr->veid );
	write_unlock_irq( &devperms_hash_guard );

#ifdef CONFIG_FAIRSCHED
	ve_fairsched_detach( ptr );
#endif
	
	write_lock_irq( &ve_list_guard );
	if( ptr->prev )
		ptr->prev->next = ptr->next;
	else
		ve_list_head = ptr->next;
	if( ptr->next )
		ptr->next->prev = ptr->prev;
	write_unlock_irq( &ve_list_guard );
	kfree( ptr );

	current->envid = envid;
	unlock_kernel();
}

static void free_tty_driver( struct tty_driver *driver )
{
	if( !driver )
		return;

	clear_termios( driver );

	if( driver->other )
		driver->other->other = NULL;
	else
		kfree( driver->driver_state );

	kfree( driver->table );
	kfree( driver->termios );
	kfree( driver->termios_locked );
	kfree( driver );
}

static struct tty_driver *alloc_tty_driver( struct tty_driver *base, void *state,
					    struct ve_struct *envid  )
{
	struct tty_driver *driver = kmalloc( sizeof(struct tty_driver), GFP_KERNEL );
	if( !driver )
		return NULL;
	
	memcpy( driver, base, sizeof(struct tty_driver));

	driver->refcount = &envid->pty_refcount;
	driver->prev = driver->next = NULL;

	driver->table          = kmalloc( sizeof(void*)*NR_PTYS, GFP_KERNEL );
	driver->termios        = kmalloc( sizeof(void*)*NR_PTYS, GFP_KERNEL );
	driver->termios_locked = kmalloc( sizeof(void*)*NR_PTYS, GFP_KERNEL );
	driver->driver_state   = state ? state :
		                 kmalloc( sizeof(struct pty_struct)*NR_PTYS, GFP_KERNEL );
	
	if( !driver->table || !driver->termios || !driver->termios_locked ||
	    !driver->driver_state )
	{
		free_tty_driver( driver );
		return NULL;
	}

	driver->envid = envid;
	driver->flags |= TTY_DRIVER_INSTALLED;
	
	memset( driver->table, 0, sizeof(void*)*NR_PTYS );
	memset( driver->termios, 0, sizeof(void*)*NR_PTYS );
	memset( driver->termios_locked, 0, sizeof(void*)*NR_PTYS );
	if( !state )
	{
		int i;
		memset( driver->driver_state, 0, sizeof(struct pty_struct)*NR_PTYS );
		for (i = 0; i < NR_PTYS; i++)
			init_waitqueue_head(&((struct pty_struct*)
					      driver->driver_state)[i].open_wait);
	}
	return driver;
}


static void register_tty_driver( struct tty_driver *driver )
{
	struct tty_driver *ptr = driver;
	while( ptr->next )
		ptr = ptr->next;
	ptr->next = tty_drivers;
	tty_drivers->prev = ptr;
	tty_drivers = driver;
}

static int  register_ve_tty_driver( struct ve_struct* envid )
{
#ifdef CONFIG_UNIX98_PTYS
	int i;
#endif
	
	/* Traditional BSD devices */
	if( !(envid->pty_driver = alloc_tty_driver( &pty_driver, NULL, envid )) )
		return -ENOMEM;
	if( !(envid->pty_slave_driver = alloc_tty_driver( &pty_slave_driver,
							  envid->pty_driver->driver_state,
							  envid )))
		return -ENOMEM;

	envid->pty_driver->other       = envid->pty_slave_driver;
	envid->pty_slave_driver->other = envid->pty_driver;

#ifdef CONFIG_UNIX98_PTYS
	for ( i = 0 ; i < UNIX98_NR_MAJORS ; i++ )
	{
		if( !(envid->ptm_driver[i] = alloc_tty_driver( &ptm_driver[i], NULL, envid )) )
			return -ENOMEM;
		if( !(envid->pts_driver[i] = alloc_tty_driver( &pts_driver[i],
							       envid->ptm_driver[i]->driver_state,
							       envid )) )
			return -ENOMEM;
		
		envid->ptm_driver[i]->other = envid->pts_driver[i];
		envid->pts_driver[i]->other = envid->ptm_driver[i];
	}

	for ( i = 0 ; i < UNIX98_NR_MAJORS ; i++ )
	{
		envid->ptm_driver[i]->next = envid->pts_driver[i];
		if( i>0 )
			envid->ptm_driver[i]->prev = envid->pts_driver[i-1];
		envid->pts_driver[i]->prev = envid->ptm_driver[i];
		if( i<UNIX98_NR_MAJORS-1 )
			envid->pts_driver[i]->next = envid->ptm_driver[i+1];
	}
#endif
	
	/* Register driver manually */
	write_lock_irq( &tty_driver_guard );
	envid->pty_driver->next = envid->pty_slave_driver;
	envid->pty_slave_driver->prev = envid->pty_driver;

#ifdef CONFIG_UNIX98_PTYS
	register_tty_driver( envid->ptm_driver[0] );
#endif
	register_tty_driver( envid->pty_driver );
	write_unlock_irq( &tty_driver_guard );

#ifdef CONFIG_DEVPTS_FS
	if( ve_register_devpts_fs( envid )<0 )
		return -ENOMEM;
#endif
	return 0;
}

static void unregister_ve_tty_driver( struct ve_struct* envid )
{
#ifdef CONFIG_UNIX98_PTYS
	int i;
#endif
	
	if( do_is_super_ve(envid) ) /* VE0 */
		return;

	ve_unregister_devpts_fs( envid );
	
	if( envid->pty_driver && envid->pty_driver->next ) /* Incomplete init ? */
	{
		struct tty_driver *ptr = envid->pty_driver;

		write_lock_irq( &tty_driver_guard );
		while( ptr->next && ptr->next->envid==envid )
			ptr = ptr->next;
		
		if (envid->pty_driver->prev)
			envid->pty_driver->prev->next = ptr->next;
		else
			tty_drivers = ptr->next;
	
		if (ptr->next)
			ptr->next->prev = envid->pty_driver->prev;
		write_unlock_irq( &tty_driver_guard );
	}

	free_tty_driver( envid->pty_driver );
	free_tty_driver( envid->pty_slave_driver );
	
#ifdef CONFIG_UNIX98_PTYS
	for ( i = 0 ; i < UNIX98_NR_MAJORS ; i++ )
	{
		free_tty_driver( envid->ptm_driver[i] );
		free_tty_driver( envid->pts_driver[i] );
	}
#endif
}


/*
 * Free the termios and termios_locked structures because
 * we don't want to get memory leaks when modular tty
 * drivers are removed from the kernel.
 */
static void clear_termios( struct tty_driver* driver )
{
	int i;
	struct termios *tp;
	if( !driver || !driver->termios )
		return;
	
	for (i = 0; i < driver->num; i++) {
		tp = driver->termios[i];
		if (tp) {
			driver->termios[i] = NULL;
			kfree(tp);
		}
		tp = driver->termios_locked[i];
		if (tp) {
			driver->termios_locked[i] = NULL;
			kfree(tp);
		}
	}
}

#ifdef CONFIG_VE_NET
struct ve_struct *get_ve_by_ip( u32 addr )
{
	struct ve_struct *ptr;

	read_lock( &ve_list_guard );
	for( ptr = ve_list_head; ptr; ptr = ptr->next )
		if( ptr->ip == addr )
			break;
	read_unlock( &ve_list_guard );
	return ptr ? ptr : child_reaper->envid;
}
#endif

#ifdef CONFIG_VE_LINK
/* 
 *  Process one VElink message.
 *  The major part of the code is taken from rtnetlink processing.
 *  This idea belongs to A. Kuznetsov
 */
#include <asm/uaccess.h>
#include <net/sock.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>

#define VE_CONF_SIZE(len) (len+sizeof(struct ve_conf_request))

pid_t daemon_pid = 0;
struct request_struct 
{
	struct list_head list;
	ulong  id;
	int    answer;
	struct semaphore waiter;
};
LIST_HEAD(request_list);
spinlock_t request_list_guard = SPIN_LOCK_UNLOCKED;

asmlinkage int sys_ve_conf_request( char *user_request, int len )
{
	struct sk_buff *skb = alloc_skb(NLMSG_GOODSIZE, GFP_KERNEL);
	struct nlmsghdr *nlh;
	struct ve_conf_request *req;
	int    err = 0;
	struct request_struct *waitq;

	if (!skb)
		return -ENOBUFS;
	
	waitq = kmalloc(sizeof(struct request_struct), GFP_KERNEL);
	if( !waitq )
	{
		err = -ENOMEM;
		goto req_alloc_failed;
	}
	memset(waitq,0,sizeof(struct request_struct));
	sema_init(&waitq->waiter,0);
	
	nlh = NLMSG_PUT(skb, daemon_pid, jiffies, VE_CONF_MSG, VE_CONF_SIZE(len));
	nlh->nlmsg_flags = NLM_F_ECHO | NLM_F_REQUEST;

	req = NLMSG_DATA(nlh);
	req->veid = current->envid->veid;
	req->id   = waitq->id = jiffies;

	if( (err=copy_from_user( req->data, user_request, len ))<0 )
		goto  nlmsg_failure;
	if( (err=netlink_unicast(nlve0, skb, daemon_pid, MSG_DONTWAIT))<0 )
		goto  nlmsg_failure;

	spin_lock_irq( &request_list_guard );
	list_add( &waitq->list, &request_list );
	spin_unlock_irq( &request_list_guard );

	printk( "awaiting request answer\n" );
	down_interruptible( &waitq->waiter );

	printk( "answer arrived %d\n", waitq->answer );
	spin_lock_irq( &request_list_guard );
	list_del( &waitq->list );
	spin_unlock_irq( &request_list_guard );
	err = waitq->answer;
	kfree(waitq);

	return err;

  nlmsg_failure:
  req_alloc_failed:
	printk( "error in sender\n" );
	kfree(waitq);
	kfree_skb(skb);
	return err;
}


extern __inline__ int
velink_rcv_msg(struct sk_buff *skb, struct nlmsghdr *nlh, int *errp)
{
	int type;
	struct ve_conf_request *req;
	struct list_head *it;

	/* Only requests from VE0 are accepted */
	if( !is_super_ve(skb) )
		return 0;

	/* Only requests are handled by kernel now */
	if (!(nlh->nlmsg_flags&NLM_F_REQUEST))
		return 0;

	type = nlh->nlmsg_type;

	/* A control message: ignore them */
	if (type < RTM_BASE)
		return 0;

	/* Unknown message: reply with EINVAL */
	if (type > RTM_MAX)
		goto err_inval;

	/* All the messages must have at least 1 byte length */
	if (nlh->nlmsg_len < NLMSG_LENGTH(sizeof(struct ve_conf_request)))
		return 0;

	/* Register user space daemon */
	if (type==VE_REGISTER) 
	{
		daemon_pid = nlh->nlmsg_pid;
		return 0;
	}

	if( type != VE_CONF_MSG )
		goto out;
	
	req = NLMSG_DATA(nlh);
	printk( "answer arrived to receiver %d\n", req->answer );
	list_for_each(it,&request_list) {
		struct request_struct *ptr = list_entry(it, struct request_struct, list);
		
		if( ptr->id==req->id )
		{
			printk( "questioner found\n" );
			ptr->answer = req->answer;
			up( &ptr->waiter );
			goto out;
		}
	}
			
out:
	printk( "all ok. quiting receiver\n" );
	return 0;

err_inval:
	*errp = -EINVAL;
	return -1;
}

extern __inline__ int velink_rcv_skb(struct sk_buff *skb)
{
	int err;
	struct nlmsghdr * nlh;

	while (skb->len >= NLMSG_SPACE(0)) {
		u32 rlen;

		nlh = (struct nlmsghdr *)skb->data;
		if (nlh->nlmsg_len < sizeof(*nlh) || skb->len < nlh->nlmsg_len)
			return 0;
		rlen = NLMSG_ALIGN(nlh->nlmsg_len);
		if (rlen > skb->len)
			rlen = skb->len;
		if (velink_rcv_msg(skb, nlh, &err)) {
			/* Not error, but we must interrupt processing here:
			 *   Note, that in this case we do not pull message
			 *   from skb, it will be processed later.
			 */
			if (err == 0)
				return -1;
			netlink_ack(skb, nlh, err);
		} else if (nlh->nlmsg_flags&NLM_F_ACK)
			netlink_ack(skb, nlh, 0);
		skb_pull(skb, rlen);
	}

	return 0;
}

/*
 *  venetlink input queue processing routine:
 *	- try to acquire shared lock. If it is failed, defer processing.
 *	- feed skbs to velink_rcv_skb, until it refuse a message,
 *	  that will occur, when a dump started and/or acquisition of
 *	  exclusive lock failed.
 */

static void velink_rcv(struct sock *sk, int len)
{
	do {
		struct sk_buff *skb;

		if (rtnl_shlock_nowait())
			return;

		while ((skb = skb_dequeue(&sk->receive_queue)) != NULL) {
			if (velink_rcv_skb(skb)) {
				if (skb->len)
					skb_queue_head(&sk->receive_queue, skb);
				else
					kfree_skb(skb);
				break;
			}
			kfree_skb(skb);
		}

		up(&rtnl_sem);
	} while (nlve0 && nlve0->receive_queue.qlen);
}
#else
asmlinkage int sys_ve_conf_request( char *user_request, int len )
{
	return -ENOSYS;
}
#endif


/* Here we creates information for 'default' VE, i.e. no VE
 * No locks required as this is a part of init and invoked
 * at the very beginning
 */
void __init init_ve_system(void)
{
	struct task_struct *init_entry = current;
	int i;
	struct ve_struct   *ptr = kmalloc( sizeof(struct ve_struct), GFP_KERNEL );

	printk( "Initializing VE sub-system\n" );
	memset( ptr, 0, sizeof(struct ve_struct) );
	ptr->init_entry = init_entry;
	cap_set_full(ptr->cap_default);

	hash_devperms( original_perms );
	hash_devperms( original_perms+1 );
	
	ve_list_head = ptr;
	sema_init( &ptr->init_exit_guard, 1 );
	atomic_set( &ptr->pcounter, 1+smp_num_cpus );

	/* Don't forget about idle tasks */
	for( i=0; i<NR_CPUS; i++ )
		if( init_tasks[i]!=NULL ) 
		{
			printk( "Marking %d idle task with zero envid\n", i );
			init_tasks[i]->envid = ptr;
		}

#ifdef CONFIG_FAIRSCHED
	fairsched_ve_init( ptr );
#endif
	
	/* Preparing default capability list */
	cap_set_full(env_cap_default);
	cap_lower(env_cap_default, CAP_SYS_MODULE);
	cap_lower(env_cap_default, CAP_SYS_RAWIO);
/*	cap_lower(env_cap_default, CAP_SYS_NICE);*/
	cap_lower(env_cap_default, CAP_SYS_TIME);
	cap_lower(env_cap_default, CAP_NET_ADMIN);

	read_lock(&init_entry->fs->lock);
	ptr->fs_rootmnt = init_entry->fs->rootmnt;
	ptr->fs_root = init_entry->fs->root;
	read_unlock(&init_entry->fs->lock);

	down_write( &uts_sem );
	memcpy( &ptr->_system_utsname, &_system_utsname, sizeof(struct new_utsname) );
	init_entry->envid = ptr;
	up_write( &uts_sem );
}

void __init postinit_ve_system(void)
{
#ifdef CONFIG_VE_LINK
	nlve0 = netlink_kernel_create(VELINK, velink_rcv);
#endif
	create_proc_read_entry("veinfo", 0, NULL, ve_read_proc, NULL);

#ifdef CONFIG_DEVPTS_FS
	ve_devpts_fs_init( get_current_veid() );
#endif
}
