/*
 *  linux/mm/kubd.c
 *
 *  Copyright (C) 2000  Andrey Moruga <moruga@sw.com.sg>
 *
 * TODO:
 *   - consider what should be done (if any) for bad pgd/pmd entries
 *
 * Changes:
 *   2000/07/28  Andrey V. Savochkin <saw@saw.sw.com.sg>
 *	- some cosmetic changes
 */

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/beancounter.h>
#include <linux/ubhash.h>
#include <linux/spinlock.h>
#include <asm/page.h>
#include <asm/pgtable.h>

/* Page weights (it's a scale for our fixed-point arithmetic).
 * Restrictions:
 *  - UB_PAGE_WEIGHT should be considerably greater than maximum number of
 *    different mm in the system (is defines the precision of calculations)
 *  - MAX_ULONGLONG/UB_PAGE_WEIGHT should be greater than the amount of
 *    physical memory (to avoid overflows)
 */
#define UB_PAGE_WEIGHT_SHIFT 24
#define UB_PAGE_WEIGHT (1 << UB_PAGE_WEIGHT_SHIFT)

/*
 *	Various debugging stuff
 *  1	main cycle
 *  2	pte
 */
#define UB_DEBUG 2

static inline void kcharge_pte (pte_t *pte, struct user_beancounter *ub)
{
	struct page *page;
	int nref;
	
	if (!pte_present(*pte))
		return;

	page = pte_page(*pte);
	if ((page-mem_map >= max_mapnr) || PageReserved(page))
		return;
	
	nref = page_count(page) - !!page->buffers 
		- (PageSwapCache(page) ? 1 : !!page->mapping);
	/* now charge the beancounter with page_weight / nreferences */
	if (nref <= 0) {
#if UB_DEBUG & 2
		printk("Error: reference counter is %d\n", nref);
		printk("Page is %p, counter %d\n", page, page_count(page));
		printk("Buffers is %p, Mapping %p\n", page->buffers,
				page->mapping);
		printk("PageSwapCache(page) is %d\n", PageSwapCache(page));
		printk("PageReserved(page) is %d\n", PageReserved(page));
		printk("page-mem_map is %d, max_mapnr %ld\n", page-mem_map,
				max_mapnr);
#endif
	} else
		ub->ub_held_pages += UB_PAGE_WEIGHT / nref;
}
	
static inline void kcharge_pmd (pmd_t *pmd, unsigned long address, 
		unsigned long end, struct user_beancounter *ub)
{
	pte_t *pte;
	unsigned long pmd_end;

	if (pmd_none(*pmd))
		return;
	if (pmd_bad(*pmd)) {
		pmd_ERROR(*pmd);
		pmd_clear(pmd);
		return;
	}
		
	pte = pte_offset(pmd, address);

	pmd_end = (address + PMD_SIZE) & PMD_MASK;
	if (end > pmd_end)
		end = pmd_end;

	do {
		kcharge_pte(pte, ub);
		address += PAGE_SIZE;
		pte++;
	} while (address && (address < end));
}
			

static inline void kcharge_pgd (pgd_t *dir, unsigned long address, 
		unsigned long end, struct user_beancounter *ub)
{
	pmd_t *pmd;
	unsigned long pgd_end;

	if (pgd_none(*dir))
		return;
	if (pgd_bad(*dir)) {
		pgd_ERROR(*dir);
		pgd_clear(dir);
		return;
	}
	pmd = pmd_offset(dir, address);

	pgd_end = (address + PGDIR_SIZE) & PGDIR_MASK;
	if (pgd_end && (end > pgd_end))
		end = pgd_end;

	do {
		kcharge_pmd(pmd, address, end, ub);

		address = (address + PMD_SIZE) & PMD_MASK;
		pmd++;
	} while (address && (address < end));
}
			
static inline void kcharge_vma (struct vm_area_struct *vma,
		struct user_beancounter *ub)
{
	unsigned long address = vma->vm_start;
	unsigned long end = vma->vm_end;
	pgd_t *pgdir;

	pgdir = pgd_offset(vma->vm_mm, address);
	do {
		/* go through pgd */
		kcharge_pgd(pgdir, address, end, ub);
		address = (address + PGDIR_SIZE) & PGDIR_MASK;
		pgdir++;
	} while (address && (address < end));
}

static inline void kcharge_mm (struct mm_struct *mm, 
		struct user_beancounter	*ub)
{
	struct vm_area_struct *vma;

	spin_lock(&mm->page_table_lock);
	for (vma = mm->mmap; vma != NULL; vma = vma->vm_next)
		kcharge_vma (vma, ub);
	spin_unlock(&mm->page_table_lock);

}

/* The background daemon, started as a kernel thread and repeatedly calculating
 * and charging the current physical memory consumption */
int kubd(void *unused)
{
	struct task_struct *tsk = current;
	struct ub_hash_slot *slot;
	struct task_struct *p;
	struct user_beancounter *ub;
	unsigned long flags;
	long i;

	tsk->session = 1;
	tsk->pgrp = 1;
	strcpy(tsk->comm, "kubd");
	sigfillset(&tsk->blocked);

repeat:
	/* first of all, clear the current ub_held_pages values */
	slot = ub_hash;
	for (i = 0; i < UB_HASH_SIZE; i++) {
		lock_beancounters(slot, flags);
		for (ub = slot->ubh_beans; ub != NULL; ub = ub->ub_next)
			ub->ub_held_pages = 0;
		unlock_beancounters(slot, flags);
		slot++;
	}

	/* now, go through all the processes -> mm -> vma -> pgd -> pmd
	 * -> page to find out how many processes reference the page */
	/* we have to keep the tasklist locked */
	read_lock(&tasklist_lock);
	for (p = init_task.next_task; p != &init_task; 
			p = p->next_task) {

		struct mm_struct *mm = p->mm;
		struct user_beancounter *ub;

		if (!p->swappable || !mm) 
			continue;
		ub = mm->beancounter;
		if (!ub)
			continue;
		kcharge_mm(mm, ub);
	}
	read_unlock(&tasklist_lock);

	/* now the calculations are done;
	 * set ub_held[UB_RESPAGES] values */
	slot = ub_hash;
	for (i = 0; i < UB_HASH_SIZE; i++) {
		lock_beancounters(slot, flags);
		for (ub = slot->ubh_beans; ub != NULL; ub = ub->ub_next) {
			ub->ub_held[UB_RESPAGES] =
				ub->ub_held_pages >> UB_PAGE_WEIGHT_SHIFT;
			if (ub->ub_maxheld[UB_RESPAGES] <
					ub->ub_held[UB_RESPAGES])
				ub->ub_maxheld[UB_RESPAGES] =
					ub->ub_held[UB_RESPAGES];
		}
		unlock_beancounters(slot, flags);
		slot++;
	}
	/* work's done, time to get sleep a bit */
	tsk->state = TASK_INTERRUPTIBLE;
#if UB_DEBUG & 1
	printk("Gonna sleep now\n");
#endif
	schedule_timeout(HZ);
#if UB_DEBUG & 1
	printk("Just woke up\n");
#endif
	goto repeat;
}

static int __init kubd_init(void)
{
	printk("Starting kubd\n");
	kernel_thread(kubd, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
	return 0;
}

module_init(kubd_init)
