
/*
 *
 *  Copyright (C) 2000  SWSoft Pte Ltd
 *
 *  Author Denis V. Lunev den@asp-linux.com
 *
 *  Quick and dirty utility for starting and stopping VEs.
 *
 *  Usage:
 *	envutil --create ENVID IP-ADDR
 *	envutil --kill	 ENVID
 *	envutil --enter  ENVID
 *	envutil --perms  ENVID ( --chr | --blk ) major minor mask
 *
 *  Changes:
 *
 *    2000/03/14 Denis Lunev <den@asp-linux.com>
 *	- created
 *    2000/06/20 Denis Lunev <den@asp-linux.com>
 *	- added devise permission
 *    2000/06/20 Denis Lunev <den@asp-linux.com>
 *	- ifconfig is now executed before startup 
 *    2000/07/19 Yuri Pudgorodsky <yur@asp-linux.com>
 *	- user benacounters support
 *    2000/07/25 Yuri Pudgorodsky <yur@asp-linux.com>
 *      - v0.0: first public release for VE 0.4 / 2.4.0-test4
 *    2000/10/17 Yuri Pudgorodsky <yur@asp-linux.com>
 *      - error messages and exit code cleanup
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <linux/kdev_t.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <string.h>
#include <linux/unistd.h>
#include <stdio.h>
#include "ve.h"


#ifdef __NR_setluid

#include <asm/page.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <linux/beancounter.h>

static _syscall1(long, setluid, uid_t, uid);
static _syscall3(long, setublimit, uid_t, uid, unsigned long, resource,
        unsigned long *, rlim);

#define PAGES_MB(mb)	((mb)*1024*1024/PAGE_SIZE)
#define MB(mb)		((mb)*1024*1024)

#define SET_UB_LIMIT(id,name,value) \
   do { \
        unsigned long limits[2]; \
	limits[0] = (value); \
	limits[1] = limits[0] + limits[0]/10; \
	if (setublimit((id), name, limits) == -1) { \
		perror("setublimit(" #name ")"); \
		exit(errno); \
	} \
   } while (0)

static void set_luid(uid_t id)
{
	if (setluid(id) == -1) {
		perror("setluid");
		exit(errno);
	}
}


static void set_ub_limits(uid_t id)
{
	if (setluid(id) == -1) {
		perror("setluid");
		exit(errno);
	}

	/* Number of processes. */
	SET_UB_LIMIT(id, UB_NUMPROC, 100);
	/* Number of sockets. */
	SET_UB_LIMIT(id, UB_NUMSOCK, 128);
	/* Number of file locks. */
	SET_UB_LIMIT(id, UB_NUMFLOCK, 1024);
	/* Number of PTYs. */
	SET_UB_LIMIT(id, UB_NUMPTY, 32);
	/* Number of siginfos. */
	SET_UB_LIMIT(id, UB_NUMSIGINFO, 256);
	/* All resident pages, for swapout guarantee. */
	SET_UB_LIMIT(id, UB_RESPAGES,  PAGES_MB(12));
	/* Address space size in pages. */
	SET_UB_LIMIT(id, UB_TOTVMPAGES,  PAGES_MB(128));
	/* OOM Killer guarantee . */
	SET_UB_LIMIT(id, UB_OOMGUARPAGES,  PAGES_MB(64));
	/* vm_enough_memory guarantee . */
	//SET_UB_LIMIT(id, UB_SPCGUARPAGES,  PAGES_MB(64));
	/* Mlock()ed pages. */
	SET_UB_LIMIT(id, UB_LOCKEDPAGES,  PAGES_MB(4));
	/* IPC SHM segment size. */
	SET_UB_LIMIT(id, UB_SHMPAGES,  PAGES_MB(8));
	/* Anonymous shared memory. */
	SET_UB_LIMIT(id, UB_ZSHMPAGES,  PAGES_MB(8));	
	/* Unswappable kernel memory size. */
	SET_UB_LIMIT(id, UB_KMEMSIZE, MB(8));
}

#else

static void set_luid(uid_t id)
{
}

static void set_ub_limits(uid_t id)
{
}

#endif

static void usage(void);
static int parse_args( int argc, char *argv[] );

static void usage(void)
{
	int envid;

#ifdef __NR_setluid
	fprintf(stderr, "envutil v0.0 with user beancounters (setluid %d, env_create %d)\n\n", __NR_setluid, __NR_env_create);
#else
	fprintf(stderr, "envutil v0.0 (env_create %d)\n\n", __NR_env_create );
#endif
	fprintf(stderr, "Usage: envutil [--create] [--kill] [--enter] [--perms]\n" );
	fprintf(stderr, "\t--create\tid addr [--bash]\n" );
	fprintf(stderr, "\t--kill\t\tid\n" );
	fprintf(stderr, "\t--enter\t\tid\n" );
	fprintf(stderr, "\t--perms\t\tid (--chr or --blk ) major minor mask\n" );
	fprintf(stderr, "\n");
	envid = env_create(0,0,0);
	if (envid < 0 && errno == ENOSYS)
		fprintf(stderr, 
			"Your kernel lacks support for virtual environments\n"
			"(compiled without CONFIG_VE)\n" );
	else if (envid < 0)
		perror("env_create");
	else {
		fprintf(stderr, "Current envid %d\n", envid );
		exit(0);
	}
	exit(errno);
}

static int parse_args( int argc, char *argv[] )
{
	int id;
	char buffer[256];
		
	if( argc<3 )
		usage();
	if( !sscanf(argv[2], "%d", &id) )
		usage();
	if( id<0 )
		usage();
	
	sprintf(buffer, "/ve_root/%d/", id );
	if( chdir(buffer) )
	{
		fprintf(stderr, "No environment root (%s), errno=%d\n", buffer, errno );
		exit(errno);
	}
	if( chroot(buffer) )
	{
		fprintf(stderr, "Can't chroot to VE root (%s), errno=%d\n", buffer, errno );
		exit(errno);
	}
	return id;
}


static char * argv_init[] = { "init", NULL };
static char * envp_init[] = { "HOME=/", "TERM=linux", NULL };
static char * argv_bash[] = { "bash", NULL };
static char * envp_bash[] = { "HOME=/", "TERM=linux", NULL };

static void empty_sigmask(void)
{
	sigset_t sigset;
	sigemptyset( &sigset );
	sigprocmask( SIG_SETMASK, &sigset, NULL );
}

int main(int argc, char *argv[]) 
{
	if( argc<2 )
		usage();
	if( !strcmp(argv[1], "--create") )
	{
		int id = parse_args( argc, argv ), child_pid;
		struct in_addr addr;
		char buffer[256];
		char buffer1[256];
		int opt_bash_init = 0;

		if( argc<4 )
			usage();
		if( !inet_aton( argv[3], &addr ))
			usage();

		strcpy( buffer1, argv[3] );
		strcpy( strrchr( buffer1, '.' )+1, "255" );
		sprintf(buffer, "/sbin/ifconfig eth0:%d %s netmask 255.255.255.0 broadcast %s",
			 id, argv[3], buffer1 );

		system(buffer);

		if (argc == 5) {
			if (!strcmp(argv[4], "--bash"))
				opt_bash_init = 1;
		}

		if( (child_pid=fork())<0 )
		{
			fprintf(stderr, "Can't fork, exiting\n" );
			exit(errno);
		} else if( child_pid==0 ) {
			int i = 0;

			set_ub_limits(id);

			if (env_create(id, VE_CREATE | VE_EXCLUSIVE, addr.s_addr ) < 0)
			{
				perror("env_create");
				return 0;
			}

			empty_sigmask();

			for( i=0; i<OPEN_MAX; i++ )
				close(i);

			if (opt_bash_init) {
				execve("/bin/bash",argv_bash,envp_bash);
				fprintf(stderr, "No /bin/bash found\n" );
			} else {
				execve("/sbin/init",argv_init,envp_init);
				execve("/etc/init",argv_init,envp_init);
				execve("/bin/init",argv_init,envp_init);
				fprintf(stderr, "No init found\n" );
			}
		}
	} else if( !strcmp(argv[1], "--kill") )
	{
		int child_pid, i;
		int id = parse_args( argc, argv );
		char buffer[256];

		sprintf(buffer, "/sbin/ifconfig eth0:%d down", id );
		system(buffer);

		if( (child_pid=fork())<0 ) {
			fprintf(stderr, "Can't fork, exiting\n" );
			exit(errno);
		} if( child_pid==0 ) { /* Child */
			int status = 0;
			char * argv_init[] = { "init", "0", NULL };
			char * envp_init[] = { "HOME=/", "TERM=linux", NULL };
					
			empty_sigmask();

			if ((status = mark_env_to_down(id)) < 0) {
				perror("Can't enter into VE");
				exit(errno);
			}
				
			if ((status = env_create(id, VE_ENTER, 0)) < 0) {
				perror("Can't enter into VE");
				exit(errno);
			}

			execve("/sbin/init",argv_init,envp_init);
			execve("/etc/init",argv_init,envp_init);
			execve("/bin/init",argv_init,envp_init);
		}
		waitpid( child_pid, &child_pid, 0 );
ok_loop:	for( i=0; i<30; i++ )
		{
			if(env_create( id, VE_TEST, 0 ) < 0 && ESRCH == errno )
			{
				fprintf(stderr, "Environment %d is downed\n", id );
				return 0;
			}
			sleep(1);
		}

		if( (child_pid=fork())<0 ) {
			fprintf(stderr, "Can't fork, exiting\n" );
			exit(errno);
		} if( child_pid==0 ) { /* Child */
			int status = 0;

			empty_sigmask();

			status = env_create( id, VE_ENTER, 0 );
			if( status==-EBUSY )
				goto ok_loop;
			if( status < 0 )
			{
				perror("Can't enter into VE");
				exit(errno);
			}
			fprintf(stderr, "Sending all VE processes TERM signal\n" );
			kill( -1, SIGTERM );
			sleep(5);
			fprintf(stderr, "Sending all VE processes KILL signal\n" );
			kill( -1, SIGKILL );
			fprintf(stderr, "Killing init\n" );
			kill( 1, SIGKILL );
			goto ok_loop;
		}
	} else if( !strcmp(argv[1], "--enter") )
	{
		int id = parse_args( argc, argv );
		int status = env_create(id, VE_ENTER, 0);

		if( id != status )
		{
			perror("Environment is not changed");
			return 0;
		}
		set_luid(id);
		fprintf(stderr, "Switching to %d\n", id );
		execve("/bin/bash",argv_bash,envp_bash);
	} else if( !strcmp(argv[1], "--perms") ) {
		int id = parse_args( argc, argv );
		unsigned minor, major, mask, type=0;

		if( argc<7 )
			usage();
		if( !sscanf(argv[5],"%u",&minor ) || !sscanf(argv[4],"%u",&major ) ||
		    !sscanf(argv[6],"%u",&mask ))
			usage();
		if( !strcmp(argv[3], "--chr") )
			type = S_IFCHR;
		else if( !strcmp(argv[3], "--blk") )
			type = S_IFBLK;
		else
			usage();
		
		fprintf(stderr, "Setting %s device (MAJOR=%d, MINOR=%d) mask=%d\n",
			S_ISCHR(type) ? "CHARACTER" : "BLOCK",
			major, minor, mask );
		id = setdevperms( id, type, MKDEV(major,minor), mask );
		fprintf(stderr, "Return status is %d\n", id );
	} else
		usage();
	return 0;
}
